@@ -62,6 +62,8 @@ struct rcxml {
 	/* keyboard */
 	int repeat_rate;
 	int repeat_delay;
+	bool kb_numlock_enable;
+	bool kb_layout_per_window;
 	struct wl_list keybinds;   /* struct keybind.link */
 
 	/* mouse */
