@@ -30,7 +30,6 @@
 #include <wlr/types/wlr_relative_pointer_v1.h>
 #include <wlr/types/wlr_pointer.h>
 #include <wlr/types/wlr_pointer_constraints_v1.h>
-#include <wlr/types/wlr_pointer_gestures_v1.h>
 #include <wlr/types/wlr_seat.h>
 #include <wlr/types/wlr_subcompositor.h>
 #include <wlr/types/wlr_xcursor_manager.h>
@@ -40,10 +39,9 @@
 #include <wlr/types/wlr_virtual_pointer_v1.h>
 #include <wlr/types/wlr_virtual_keyboard_v1.h>
 #include <wlr/util/log.h>
-#include <xkbcommon/xkbcommon.h>
-#include "cursor.h"
 #include "config/keybind.h"
 #include "config/rcxml.h"
+#include "input/cursor.h"
 #include "regions.h"
 #include "session-lock.h"
 #if HAVE_NLS
@@ -57,14 +55,6 @@
 #define XCURSOR_DEFAULT "left_ptr"
 #define XCURSOR_SIZE 24
 
-#ifndef MIN
-#define MIN(a, b) (((a) < (b)) ? (a) : (b))
-#endif
-
-#ifndef MAX
-#define MAX(a, b) (((a) > (b)) ? (a) : (b))
-#endif
-
 enum input_mode {
 	LAB_INPUT_STATE_PASSTHROUGH = 0,
 	LAB_INPUT_STATE_MOVE,
@@ -76,7 +66,7 @@ struct input {
 	struct wlr_input_device *wlr_input_device;
 	struct seat *seat;
 	struct wl_listener destroy;
-	struct wl_list link; /* seat::inputs */
+	struct wl_list link; /* seat.inputs */
 };
 
 /*
@@ -167,6 +157,7 @@ struct seat {
 	struct wl_client *active_client_while_inhibited;
 	struct wl_list inputs;
 	struct wl_listener new_input;
+	struct wl_listener focus_change;
 
 	struct wl_listener cursor_motion;
 	struct wl_listener cursor_motion_absolute;
@@ -245,8 +236,17 @@ struct server {
 	struct wlr_box grab_box;
 	uint32_t resize_edges;
 
-	/* SSD state */
+	/*
+	 * Currently focused view. Updated with each "focus change"
+	 * event. This view is drawn with "active" SSD coloring.
+	 */
 	struct view *focused_view;
+	/*
+	 * Most recently raised view. Used to avoid unnecessarily
+	 * raising the same view over and over.
+	 */
+	struct view *last_raised_view;
+
 	struct ssd_hover_state *ssd_hover_state;
 
 	/* Tree for all non-layer xdg/xwayland-shell surfaces */
@@ -315,12 +315,13 @@ struct server {
 	struct theme *theme;
 
 	struct menu *menu_current;
+	struct wl_list menus;
 };
 
 #define LAB_NR_LAYERS (4)
 
 struct output {
-	struct wl_list link; /* server::outputs */
+	struct wl_list link; /* server.outputs */
 	struct server *server;
 	struct wlr_output *wlr_output;
 	struct wlr_scene_output *scene_output;
@@ -350,10 +351,7 @@ struct constraint {
 };
 
 void xdg_popup_create(struct view *view, struct wlr_xdg_popup *wlr_popup);
-
-void xdg_activation_handle_request(struct wl_listener *listener, void *data);
-
-void xdg_surface_new(struct wl_listener *listener, void *data);
+void xdg_shell_init(struct server *server);
 
 void foreign_toplevel_handle_create(struct view *view);
 void foreign_toplevel_update_outputs(struct view *view);
@@ -370,10 +368,36 @@ void foreign_toplevel_update_outputs(struct view *view);
  *              cannot assume this means that the window actually has keyboard
  *              or pointer focus, in this compositor are they called together.
  */
-void desktop_focus_and_activate_view(struct seat *seat, struct view *view);
+
+/**
+ * desktop_focus_view() - do multiple things to make a view "active" and
+ * ready to use:
+ *  - unminimize
+ *  - switch to the workspace it's on
+ *  - give input (keyboard) focus
+ *  - optionally raise above other views
+ *
+ * It's okay to call this function even if the view isn't mapped or the
+ * session is locked/input is inhibited; it will simply do nothing.
+ */
+void desktop_focus_view(struct view *view, bool raise);
+
+/**
+ * desktop_focus_view_or_surface() - like desktop_focus_view() but can
+ * also focus other (e.g. xwayland-unmanaged) surfaces
+ */
+void desktop_focus_view_or_surface(struct seat *seat, struct view *view,
+	struct wlr_surface *surface, bool raise);
+
 void desktop_arrange_all_views(struct server *server);
 void desktop_focus_output(struct output *output);
-struct view *desktop_topmost_mapped_view(struct server *server);
+struct view *desktop_topmost_focusable_view(struct server *server);
+
+/**
+ * Toggles the (output local) visibility of the layershell top layer
+ * based on the existence of a fullscreen window on the current workspace.
+ */
+void desktop_update_top_layer_visiblity(struct server *server);
 
 enum lab_cycle_dir {
 	LAB_CYCLE_DIR_NONE,
@@ -388,23 +412,28 @@ enum lab_cycle_dir {
  */
 struct view *desktop_cycle_view(struct server *server, struct view *start_view,
 	enum lab_cycle_dir dir);
-struct view *desktop_focused_view(struct server *server);
-void desktop_focus_topmost_mapped_view(struct server *server);
 
-void keyboard_cancel_keybind_repeat(struct keyboard *keyboard);
-void keyboard_key_notify(struct wl_listener *listener, void *data);
-void keyboard_modifiers_notify(struct wl_listener *listener, void *data);
-void keyboard_init(struct seat *seat);
-bool keyboard_any_modifiers_pressed(struct wlr_keyboard *keyboard);
-void keyboard_finish(struct seat *seat);
-
-void touch_init(struct seat *seat);
-void touch_finish(struct seat *seat);
+/**
+ * desktop_focus_topmost_view() - focus the topmost view on the current
+ * workspace, skipping views that claim not to want focus (those can
+ * still be focused by explicit request, e.g. by clicking in them).
+ *
+ * This function is typically called when the focused view is hidden
+ * (closes, is minimized, etc.) to focus the "next" view underneath.
+ */
+void desktop_focus_topmost_view(struct server *server);
 
 void seat_init(struct server *server);
 void seat_finish(struct server *server);
 void seat_reconfigure(struct server *server);
 void seat_focus_surface(struct seat *seat, struct wlr_surface *surface);
+
+/**
+ * seat_focus_lock_surface() - ONLY to be called from session-lock.c to
+ * focus lock screen surfaces. Use seat_focus_surface() otherwise.
+ */
+void seat_focus_lock_surface(struct seat *seat, struct wlr_surface *surface);
+
 void seat_set_focus_layer(struct seat *seat, struct wlr_layer_surface_v1 *layer);
 void seat_set_pressed(struct seat *seat, struct view *view,
 	struct wlr_scene_node *node, struct wlr_surface *surface,
@@ -426,6 +455,7 @@ bool output_is_usable(struct output *output);
 void output_update_usable_area(struct output *output);
 void output_update_all_usable_areas(struct server *server, bool layout_changed);
 struct wlr_box output_usable_area_in_layout_coords(struct output *output);
+struct wlr_box output_usable_area_scaled(struct output *output);
 void handle_output_power_manager_set_mode(struct wl_listener *listener,
 	void *data);
 
