@@ -7,6 +7,10 @@
 #include <stdint.h>
 #include <wayland-util.h>
 #include <wlr/util/box.h>
+#include <xkbcommon/xkbcommon.h>
+
+#define LAB_MIN_VIEW_WIDTH  100
+#define LAB_MIN_VIEW_HEIGHT  60
 
 /*
  * In labwc, a view is a container for surfaces which can be moved around by
@@ -26,6 +30,18 @@ enum ssd_preference {
 	LAB_SSD_PREF_SERVER,
 };
 
+/**
+ * Directions in which a view can be maximized. "None" is used
+ * internally to mean "not maximized" but is not valid in rc.xml.
+ * Therefore when parsing rc.xml, "None" means "Invalid".
+ */
+enum view_axis {
+	VIEW_AXIS_NONE = 0,
+	VIEW_AXIS_HORIZONTAL = (1 << 0),
+	VIEW_AXIS_VERTICAL = (1 << 1),
+	VIEW_AXIS_BOTH = (VIEW_AXIS_HORIZONTAL | VIEW_AXIS_VERTICAL),
+};
+
 enum view_edge {
 	VIEW_EDGE_INVALID = 0,
 
@@ -36,7 +52,34 @@ enum view_edge {
 	VIEW_EDGE_CENTER,
 };
 
+enum view_wants_focus {
+	/* View does not want focus */
+	VIEW_WANTS_FOCUS_NEVER = 0,
+	/* View wants focus */
+	VIEW_WANTS_FOCUS_ALWAYS,
+	/*
+	 * View should be offered focus and may accept or decline
+	 * (a.k.a. ICCCM Globally Active input model). Labwc generally
+	 * avoids focusing these views automatically (e.g. when another
+	 * view on top is closed) but they may be focused by user action
+	 * (e.g. mouse click).
+	 */
+	VIEW_WANTS_FOCUS_OFFER,
+};
+
 struct view;
+struct wlr_surface;
+
+/* Basic size hints (subset of XSizeHints from X11) */
+struct view_size_hints {
+	int min_width;
+	int min_height;
+	int width_inc;
+	int height_inc;
+	int base_width;
+	int base_height;
+};
+
 struct view_impl {
 	void (*configure)(struct view *view, struct wlr_box geo);
 	void (*close)(struct view *view);
@@ -57,7 +100,11 @@ struct view_impl {
 	void (*move_to_back)(struct view *view);
 	struct view *(*get_root)(struct view *self);
 	void (*append_children)(struct view *self, struct wl_array *children);
-	void (*fill_size_hints)(struct view *self, struct wlr_box *box);
+	/* determines if view and surface are owned by the same process */
+	bool (*is_related)(struct view *self, struct wlr_surface *surface);
+	struct view_size_hints (*get_size_hints)(struct view *self);
+	/* if not implemented, VIEW_WANTS_FOCUS_ALWAYS is assumed */
+	enum view_wants_focus (*wants_focus)(struct view *self);
 };
 
 struct view {
@@ -93,10 +140,11 @@ struct view {
 	bool ssd_titlebar_hidden;
 	enum ssd_preference ssd_preference;
 	bool minimized;
-	bool maximized;
+	enum view_axis maximized;
 	bool fullscreen;
-	uint32_t tiled;  /* private, enum view_edge in src/view.c */
+	enum view_edge tiled;
 	bool inhibits_keybinds;
+	xkb_layout_index_t keyboard_layout;
 
 	/* Pointer to an output owned struct region, may be NULL */
 	struct region *tiled_region;
@@ -159,6 +207,12 @@ struct view {
 	struct wl_listener set_title;
 };
 
+struct view_query {
+	struct wl_list link;
+	char *identifier;
+	char *title;
+};
+
 struct xdg_toplevel_view {
 	struct view base;
 	struct wlr_xdg_surface *xdg_surface;
@@ -168,13 +222,48 @@ struct xdg_toplevel_view {
 	struct wl_listener new_popup;
 };
 
+/* All criteria is applied in AND logic */
 enum lab_view_criteria {
+	/* No filter -> all focusable views */
 	LAB_VIEW_CRITERIA_NONE = 0,
-	LAB_VIEW_CRITERIA_CURRENT_WORKSPACE = 1 << 0,
-	LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP = 1 << 1,
-	LAB_VIEW_CRITERIA_NO_SKIP_WINDOW_SWITCHER = 1 << 2,
+
+	/*
+	 * Includes always-on-top views, e.g.
+	 * what is visible on the current workspace
+	 */
+	LAB_VIEW_CRITERIA_CURRENT_WORKSPACE       = 1 << 0,
+
+	/* Positive criteria */
+	LAB_VIEW_CRITERIA_FULLSCREEN              = 1 << 1,
+	LAB_VIEW_CRITERIA_ALWAYS_ON_TOP           = 1 << 2,
+
+	/* Negative criteria */
+	LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP        = 1 << 6,
+	LAB_VIEW_CRITERIA_NO_SKIP_WINDOW_SWITCHER = 1 << 7,
 };
 
+/**
+ * view_from_wlr_surface() - returns the view associated with a
+ * wlr_surface, or NULL if the surface has no associated view.
+ */
+struct view *view_from_wlr_surface(struct wlr_surface *surface);
+
+/**
+ * view_query_free() - Free a given view query
+ * @query: Query to be freed.
+ */
+void view_query_free(struct view_query *view);
+
+/**
+ * view_matches_query() - Check if view matches the given criteria
+ * @view: View to checked.
+ * @query: Criteria to match against.
+ *
+ * Returns true if %view matches all of the criteria given in %query, false
+ * otherwise.
+ */
+bool view_matches_query(struct view *view, struct view_query *query);
+
 /**
  * for_each_view() - iterate over all views which match criteria
  * @view: Iterator.
@@ -229,8 +318,18 @@ struct view *view_next(struct wl_list *head, struct view *view,
 void view_array_append(struct server *server, struct wl_array *views,
 	enum lab_view_criteria criteria);
 
+enum view_wants_focus view_wants_focus(struct view *view);
+
 /**
- * view_isfocusable() - Check whether or not a view can be focused
+ * view_is_focusable_from() - variant of view_is_focusable()
+ * that takes into account the previously focused surface
+ * @view: view to be checked
+ * @prev_surface: previously focused surface
+ */
+bool view_is_focusable_from(struct view *view, struct wlr_surface *prev);
+
+/**
+ * view_is_focusable() - Check whether or not a view can be focused
  * @view: view to be checked
  *
  * The purpose of this test is to filter out views (generally Xwayland) which
@@ -241,12 +340,14 @@ void view_array_append(struct server *server, struct wl_array *views,
  * The only views that are allowed to be focusd are those that have a surface
  * and have been mapped at some point since creation.
  */
-bool view_isfocusable(struct view *view);
+static inline bool
+view_is_focusable(struct view *view) {
+	return view_is_focusable_from(view, NULL);
+}
 
-bool view_inhibits_keybinds(struct view *view);
 void view_toggle_keybinds(struct view *view);
 
-void view_set_activated(struct view *view);
+void view_set_activated(struct view *view, bool activated);
 void view_set_output(struct view *view, struct output *output);
 void view_close(struct view *view);
 
@@ -263,8 +364,12 @@ void view_resize_relative(struct view *view,
 	int left, int right, int top, int bottom);
 void view_move_relative(struct view *view, int x, int y);
 void view_move(struct view *view, int x, int y);
+void view_move_to_cursor(struct view *view);
 void view_moved(struct view *view);
 void view_minimize(struct view *view, bool minimized);
+bool view_compute_centered_position(struct view *view,
+	const struct wlr_box *ref, int w, int h, int *x, int *y);
+bool view_adjust_floating_geometry(struct view *view, struct wlr_box *geometry);
 void view_store_natural_geometry(struct view *view);
 
 /**
@@ -276,13 +381,14 @@ void view_store_natural_geometry(struct view *view);
 void view_center(struct view *view, const struct wlr_box *ref);
 void view_restore_to(struct view *view, struct wlr_box geometry);
 void view_set_untiled(struct view *view);
-void view_maximize(struct view *view, bool maximize,
+void view_maximize(struct view *view, enum view_axis axis,
 	bool store_natural_geometry);
 void view_set_fullscreen(struct view *view, bool fullscreen);
-void view_toggle_maximize(struct view *view);
+void view_toggle_maximize(struct view *view, enum view_axis axis);
 void view_toggle_decorations(struct view *view);
 
 bool view_is_always_on_top(struct view *view);
+bool view_is_always_on_bottom(struct view *view);
 void view_toggle_always_on_top(struct view *view);
 void view_toggle_always_on_bottom(struct view *view);
 
@@ -292,8 +398,11 @@ void view_move_to_workspace(struct view *view, struct workspace *workspace);
 void view_set_decorations(struct view *view, bool decorations);
 void view_toggle_fullscreen(struct view *view);
 void view_adjust_for_layout_change(struct view *view);
-void view_move_to_edge(struct view *view, enum view_edge direction);
-void view_snap_to_edge(struct view *view, enum view_edge direction, bool store_natural_geometry);
+void view_move_to_edge(struct view *view, enum view_edge direction, bool snap_to_windows);
+void view_grow_to_edge(struct view *view, enum view_edge direction);
+void view_shrink_to_edge(struct view *view, enum view_edge direction);
+void view_snap_to_edge(struct view *view, enum view_edge direction,
+	bool across_outputs, bool store_natural_geometry);
 void view_snap_to_region(struct view *view, struct region *region, bool store_natural_geometry);
 
 void view_move_to_front(struct view *view);
@@ -301,17 +410,26 @@ void view_move_to_back(struct view *view);
 struct view *view_get_root(struct view *view);
 void view_append_children(struct view *view, struct wl_array *children);
 
+/**
+ * view_is_related() - determine if view and surface are owned by the
+ * same application/process. Currently only implemented for xwayland
+ * views/surfaces.
+ */
+bool view_is_related(struct view *view, struct wlr_surface *surface);
+
 const char *view_get_string_prop(struct view *view, const char *prop);
 void view_update_title(struct view *view);
 void view_update_app_id(struct view *view);
 void view_reload_ssd(struct view *view);
 
+struct view_size_hints view_get_size_hints(struct view *view);
 void view_adjust_size(struct view *view, int *w, int *h);
 
 void view_evacuate_region(struct view *view);
 void view_on_output_destroy(struct view *view);
 void view_destroy(struct view *view);
 
+enum view_axis view_axis_parse(const char *direction);
 enum view_edge view_edge_parse(const char *direction);
 
 /* xdg.c */
