@@ -7,6 +7,7 @@
 #include <unistd.h>
 #include <wlr/util/log.h>
 #include "action.h"
+#include "common/macros.h"
 #include "common/list.h"
 #include "common/mem.h"
 #include "common/parse-bool.h"
@@ -24,6 +25,8 @@ enum action_arg_type {
 	LAB_ACTION_ARG_STR = 0,
 	LAB_ACTION_ARG_BOOL,
 	LAB_ACTION_ARG_INT,
+	LAB_ACTION_ARG_QUERY_LIST,
+	LAB_ACTION_ARG_ACTION_LIST,
 };
 
 struct action_arg {
@@ -48,6 +51,11 @@ struct action_arg_int {
 	int value;
 };
 
+struct action_arg_list {
+	struct action_arg base;
+	struct wl_list value;
+};
+
 enum action_type {
 	ACTION_TYPE_INVALID = 0,
 	ACTION_TYPE_NONE,
@@ -58,6 +66,8 @@ enum action_type {
 	ACTION_TYPE_EXIT,
 	ACTION_TYPE_MOVE_TO_EDGE,
 	ACTION_TYPE_SNAP_TO_EDGE,
+	ACTION_TYPE_GROW_TO_EDGE,
+	ACTION_TYPE_SHRINK_TO_EDGE,
 	ACTION_TYPE_NEXT_WINDOW,
 	ACTION_TYPE_PREVIOUS_WINDOW,
 	ACTION_TYPE_RECONFIGURE,
@@ -69,6 +79,7 @@ enum action_type {
 	ACTION_TYPE_TOGGLE_ALWAYS_ON_TOP,
 	ACTION_TYPE_TOGGLE_ALWAYS_ON_BOTTOM,
 	ACTION_TYPE_FOCUS,
+	ACTION_TYPE_UNFOCUS,
 	ACTION_TYPE_ICONIFY,
 	ACTION_TYPE_MOVE,
 	ACTION_TYPE_RAISE,
@@ -76,12 +87,15 @@ enum action_type {
 	ACTION_TYPE_RESIZE,
 	ACTION_TYPE_RESIZE_RELATIVE,
 	ACTION_TYPE_MOVETO,
+	ACTION_TYPE_MOVETO_CURSOR,
 	ACTION_TYPE_MOVE_RELATIVE,
 	ACTION_TYPE_SEND_TO_DESKTOP,
 	ACTION_TYPE_GO_TO_DESKTOP,
 	ACTION_TYPE_SNAP_TO_REGION,
 	ACTION_TYPE_TOGGLE_KEYBINDS,
 	ACTION_TYPE_FOCUS_OUTPUT,
+	ACTION_TYPE_IF,
+	ACTION_TYPE_FOR_EACH,
 };
 
 const char *action_names[] = {
@@ -94,6 +108,8 @@ const char *action_names[] = {
 	"Exit",
 	"MoveToEdge",
 	"SnapToEdge",
+	"GrowToEdge",
+	"ShrinkToEdge",
 	"NextWindow",
 	"PreviousWindow",
 	"Reconfigure",
@@ -105,6 +121,7 @@ const char *action_names[] = {
 	"ToggleAlwaysOnTop",
 	"ToggleAlwaysOnBottom",
 	"Focus",
+	"Unfocus",
 	"Iconify",
 	"Move",
 	"Raise",
@@ -112,12 +129,15 @@ const char *action_names[] = {
 	"Resize",
 	"ResizeRelative",
 	"MoveTo",
+	"MoveToCursor",
 	"MoveRelative",
 	"SendToDesktop",
 	"GoToDesktop",
 	"SnapToRegion",
 	"ToggleKeybinds",
 	"FocusOutput",
+	"If",
+	"ForEach",
 	NULL
 };
 
@@ -158,6 +178,79 @@ action_arg_add_int(struct action *action, const char *key, int value)
 	wl_list_append(&action->args, &arg->base.link);
 }
 
+static void
+action_arg_add_list(struct action *action, const char *key, enum action_arg_type type)
+{
+	assert(action);
+	assert(key);
+	struct action_arg_list *arg = znew(*arg);
+	arg->base.type = type;
+	arg->base.key = xstrdup(key);
+	wl_list_init(&arg->value);
+	wl_list_append(&action->args, &arg->base.link);
+}
+
+void
+action_arg_add_querylist(struct action *action, const char *key)
+{
+	action_arg_add_list(action, key, LAB_ACTION_ARG_QUERY_LIST);
+}
+
+void
+action_arg_add_actionlist(struct action *action, const char *key)
+{
+	action_arg_add_list(action, key, LAB_ACTION_ARG_ACTION_LIST);
+}
+
+static void *
+action_get_arg(struct action *action, const char *key, enum action_arg_type type)
+{
+	assert(action);
+	assert(key);
+	struct action_arg *arg;
+	wl_list_for_each(arg, &action->args, link) {
+		if (!strcasecmp(key, arg->key) && arg->type == type) {
+			return arg;
+		}
+	}
+	return NULL;
+}
+
+static const char *
+action_get_str(struct action *action, const char *key, const char *default_value)
+{
+	struct action_arg_str *arg = action_get_arg(action, key, LAB_ACTION_ARG_STR);
+	return arg ? arg->value : default_value;
+}
+
+static bool
+action_get_bool(struct action *action, const char *key, bool default_value)
+{
+	struct action_arg_bool *arg = action_get_arg(action, key, LAB_ACTION_ARG_BOOL);
+	return arg ? arg->value : default_value;
+}
+
+static int
+action_get_int(struct action *action, const char *key, int default_value)
+{
+	struct action_arg_int *arg = action_get_arg(action, key, LAB_ACTION_ARG_INT);
+	return arg ? arg->value : default_value;
+}
+
+struct wl_list *
+action_get_querylist(struct action *action, const char *key)
+{
+	struct action_arg_list *arg = action_get_arg(action, key, LAB_ACTION_ARG_QUERY_LIST);
+	return arg ? &arg->value : NULL;
+}
+
+struct wl_list *
+action_get_actionlist(struct action *action, const char *key)
+{
+	struct action_arg_list *arg = action_get_arg(action, key, LAB_ACTION_ARG_ACTION_LIST);
+	return arg ? &arg->value : NULL;
+}
+
 void
 action_arg_from_xml_node(struct action *action, const char *nodename, const char *content)
 {
@@ -179,7 +272,14 @@ action_arg_from_xml_node(struct action *action, const char *nodename, const char
 		}
 		break;
 	case ACTION_TYPE_MOVE_TO_EDGE:
+		if (!strcasecmp(argument, "snapWindows")) {
+			action_arg_add_bool(action, argument, parse_bool(content, true));
+			goto cleanup;
+		}
+		/* Falls through */
 	case ACTION_TYPE_SNAP_TO_EDGE:
+	case ACTION_TYPE_GROW_TO_EDGE:
+	case ACTION_TYPE_SHRINK_TO_EDGE:
 		if (!strcmp(argument, "direction")) {
 			enum view_edge edge = view_edge_parse(content);
 			if ((edge == VIEW_EDGE_CENTER && action->type != ACTION_TYPE_SNAP_TO_EDGE)
@@ -198,6 +298,19 @@ action_arg_from_xml_node(struct action *action, const char *nodename, const char
 			goto cleanup;
 		}
 		break;
+	case ACTION_TYPE_TOGGLE_MAXIMIZE:
+	case ACTION_TYPE_MAXIMIZE:
+		if (!strcmp(argument, "direction")) {
+			enum view_axis axis = view_axis_parse(content);
+			if (axis == VIEW_AXIS_NONE) {
+				wlr_log(WLR_ERROR, "Invalid argument for action %s: '%s' (%s)",
+					action_names[action->type], argument, content);
+			} else {
+				action_arg_add_int(action, argument, axis);
+			}
+			goto cleanup;
+		}
+		break;
 	case ACTION_TYPE_RESIZE_RELATIVE:
 		if (!strcmp(argument, "left") || !strcmp(argument, "right") ||
 				!strcmp(argument, "top") || !strcmp(argument, "bottom")) {
@@ -249,82 +362,6 @@ cleanup:
 	free(argument);
 }
 
-static const char *
-action_str_from_arg(struct action_arg *arg)
-{
-	assert(arg->type == LAB_ACTION_ARG_STR);
-	return ((struct action_arg_str *)arg)->value;
-}
-
-static bool
-arg_value_exists(struct action *action, const char *key, enum action_arg_type type)
-{
-	assert(action);
-	assert(key);
-	struct action_arg *arg;
-	wl_list_for_each(arg, &action->args, link) {
-		if (!strcasecmp(key, arg->key) && arg->type == type) {
-			return true;
-		}
-	}
-	return false;
-}
-
-static const char *
-get_arg_value_str(struct action *action, const char *key, const char *default_value)
-{
-	assert(action);
-	assert(key);
-	struct action_arg *arg;
-	wl_list_for_each(arg, &action->args, link) {
-		if (!strcasecmp(key, arg->key)) {
-			return action_str_from_arg(arg);
-		}
-	}
-	return default_value;
-}
-
-static bool
-get_arg_value_bool(struct action *action, const char *key, bool default_value)
-{
-	assert(action);
-	assert(key);
-	struct action_arg *arg;
-	wl_list_for_each(arg, &action->args, link) {
-		if (!strcasecmp(key, arg->key)) {
-			assert(arg->type == LAB_ACTION_ARG_BOOL);
-			return ((struct action_arg_bool *)arg)->value;
-		}
-	}
-	return default_value;
-}
-
-static int
-get_arg_value_int(struct action *action, const char *key, int default_value)
-{
-	assert(action);
-	assert(key);
-	struct action_arg *arg;
-	wl_list_for_each(arg, &action->args, link) {
-		if (!strcasecmp(key, arg->key)) {
-			assert(arg->type == LAB_ACTION_ARG_INT);
-			return ((struct action_arg_int *)arg)->value;
-		}
-	}
-	return default_value;
-}
-
-static struct action_arg *
-action_get_first_arg(struct action *action)
-{
-	struct action_arg *arg;
-	struct wl_list *item = action->args.next;
-	if (item == &action->args) {
-		return NULL;
-	}
-	return wl_container_of(item, arg, link);
-}
-
 static enum action_type
 action_type_from_str(const char *action_name)
 {
@@ -368,6 +405,20 @@ actions_contain_toggle_keybinds(struct wl_list *action_list)
 	return false;
 }
 
+static bool
+action_list_is_valid(struct wl_list *actions)
+{
+	assert(actions);
+
+	struct action *action;
+	wl_list_for_each(action, actions, link) {
+		if (!action_is_valid(action)) {
+			return false;
+		}
+	}
+	return true;
+}
+
 /* Checks for *required* arguments */
 bool
 action_is_valid(struct action *action)
@@ -381,6 +432,8 @@ action_is_valid(struct action *action)
 		break;
 	case ACTION_TYPE_MOVE_TO_EDGE:
 	case ACTION_TYPE_SNAP_TO_EDGE:
+	case ACTION_TYPE_GROW_TO_EDGE:
+	case ACTION_TYPE_SHRINK_TO_EDGE:
 		arg_name = "direction";
 		arg_type = LAB_ACTION_ARG_INT;
 		break;
@@ -397,12 +450,25 @@ action_is_valid(struct action *action)
 	case ACTION_TYPE_FOCUS_OUTPUT:
 		arg_name = "output";
 		break;
+	case ACTION_TYPE_IF:
+	case ACTION_TYPE_FOR_EACH:
+		; /* works around "a label can only be part of a statement" */
+		static const char * const branches[] = { "then", "else" };
+		for (size_t i = 0; i < ARRAY_SIZE(branches); i++) {
+			struct wl_list *children = action_get_actionlist(action, branches[i]);
+			if (children && !action_list_is_valid(children)) {
+				wlr_log(WLR_ERROR, "Invalid action in %s '%s' branch",
+					action_names[action->type], branches[i]);
+				return false;
+			}
+		}
+		return true;
 	default:
 		/* No arguments required */
 		return true;
 	}
 
-	if (arg_value_exists(action, arg_name, arg_type)) {
+	if (action_get_arg(action, arg_name, arg_type)) {
 		return true;
 	}
 
@@ -422,6 +488,15 @@ action_free(struct action *action)
 		if (arg->type == LAB_ACTION_ARG_STR) {
 			struct action_arg_str *str_arg = (struct action_arg_str *)arg;
 			zfree(str_arg->value);
+		} else if (arg->type == LAB_ACTION_ARG_ACTION_LIST) {
+			struct action_arg_list *list_arg = (struct action_arg_list *)arg;
+			action_list_free(&list_arg->value);
+		} else if (arg->type == LAB_ACTION_ARG_QUERY_LIST) {
+			struct action_arg_list *list_arg = (struct action_arg_list *)arg;
+			struct view_query *elm, *next;
+			wl_list_for_each_safe(elm, next, &list_arg->value, link) {
+				view_query_free(elm);
+			}
 		}
 		zfree(arg);
 	}
@@ -448,7 +523,7 @@ show_menu(struct server *server, struct view *view, const char *menu_name)
 	}
 
 	bool force_menu_top_left = false;
-	struct menu *menu = menu_get_by_id(menu_name);
+	struct menu *menu = menu_get_by_id(server, menu_name);
 	if (!menu) {
 		return;
 	}
@@ -503,7 +578,32 @@ view_for_action(struct view *activator, struct server *server,
 		return ctx.view;
 	}
 	default:
-		return desktop_focused_view(server);
+		return server->focused_view;
+	}
+}
+
+static void
+run_if_action(struct view *view, struct server *server, struct action *action)
+{
+	struct view_query *query;
+	struct wl_list *queries, *actions;
+	const char *branch = "then";
+
+	queries = action_get_querylist(action, "query");
+	if (queries) {
+		branch = "else";
+		/* All queries are OR'ed */
+		wl_list_for_each(query, queries, link) {
+			if (view_matches_query(view, query)) {
+				branch = "then";
+				break;
+			}
+		}
+	}
+
+	actions = action_get_actionlist(action, branch);
+	if (actions) {
+		actions_run(view, server, actions, 0);
 	}
 }
 
@@ -518,19 +618,9 @@ actions_run(struct view *activator, struct server *server,
 
 	struct view *view;
 	struct action *action;
-	struct action_arg *arg;
 	wl_list_for_each(action, actions, link) {
-		/* Get arg now so we don't have to repeat every time we only need one */
-		arg = action_get_first_arg(action);
-
-		if (arg && arg->type == LAB_ACTION_ARG_STR) {
-			wlr_log(WLR_DEBUG, "Handling action %u: %s %s",
-				action->type, action_names[action->type],
-				action_str_from_arg(arg));
-		} else {
-			wlr_log(WLR_DEBUG, "Handling action %u: %s",
-				action->type, action_names[action->type]);
-		}
+		wlr_log(WLR_DEBUG, "Handling action %u: %s", action->type,
+			action_names[action->type]);
 
 		/*
 		 * Refetch view because it may have been changed due to the
@@ -563,7 +653,7 @@ actions_run(struct view *activator, struct server *server,
 			{
 				struct buf cmd;
 				buf_init(&cmd);
-				buf_add(&cmd, action_str_from_arg(arg));
+				buf_add(&cmd, action_get_str(action, "command", NULL));
 				buf_expand_tilde(&cmd);
 				spawn_async_no_shell(cmd.buf);
 				free(cmd.buf);
@@ -575,15 +665,32 @@ actions_run(struct view *activator, struct server *server,
 		case ACTION_TYPE_MOVE_TO_EDGE:
 			if (view) {
 				/* Config parsing makes sure that direction is a valid direction */
-				enum view_edge edge = get_arg_value_int(action, "direction", 0);
-				view_move_to_edge(view, edge);
+				enum view_edge edge = action_get_int(action, "direction", 0);
+				bool snap_to_windows = action_get_bool(action, "snapWindows", true);
+				view_move_to_edge(view, edge, snap_to_windows);
 			}
 			break;
 		case ACTION_TYPE_SNAP_TO_EDGE:
 			if (view) {
 				/* Config parsing makes sure that direction is a valid direction */
-				enum view_edge edge = get_arg_value_int(action, "direction", 0);
-				view_snap_to_edge(view, edge, /*store_natural_geometry*/ true);
+				enum view_edge edge = action_get_int(action, "direction", 0);
+				view_snap_to_edge(view, edge,
+					/*across_outputs*/ true,
+					/*store_natural_geometry*/ true);
+			}
+			break;
+		case ACTION_TYPE_GROW_TO_EDGE:
+			if (view) {
+				/* Config parsing makes sure that direction is a valid direction */
+				enum view_edge edge = action_get_int(action, "direction", 0);
+				view_grow_to_edge(view, edge);
+			}
+			break;
+		case ACTION_TYPE_SHRINK_TO_EDGE:
+			if (view) {
+				/* Config parsing makes sure that direction is a valid direction */
+				enum view_edge edge = action_get_int(action, "direction", 0);
+				view_shrink_to_edge(view, edge);
 			}
 			break;
 		case ACTION_TYPE_NEXT_WINDOW:
@@ -600,16 +707,21 @@ actions_run(struct view *activator, struct server *server,
 			kill(getpid(), SIGHUP);
 			break;
 		case ACTION_TYPE_SHOW_MENU:
-			show_menu(server, view, action_str_from_arg(arg));
+			show_menu(server, view, action_get_str(action, "menu", NULL));
 			break;
 		case ACTION_TYPE_TOGGLE_MAXIMIZE:
 			if (view) {
-				view_toggle_maximize(view);
+				enum view_axis axis = action_get_int(action,
+					"direction", VIEW_AXIS_BOTH);
+				view_toggle_maximize(view, axis);
 			}
 			break;
 		case ACTION_TYPE_MAXIMIZE:
 			if (view) {
-				view_maximize(view, true, /*store_natural_geometry*/ true);
+				enum view_axis axis = action_get_int(action,
+					"direction", VIEW_AXIS_BOTH);
+				view_maximize(view, axis,
+					/*store_natural_geometry*/ true);
 			}
 			break;
 		case ACTION_TYPE_TOGGLE_FULLSCREEN:
@@ -634,9 +746,12 @@ actions_run(struct view *activator, struct server *server,
 			break;
 		case ACTION_TYPE_FOCUS:
 			if (view) {
-				desktop_focus_and_activate_view(&server->seat, view);
+				desktop_focus_view(view, /*raise*/ false);
 			}
 			break;
+		case ACTION_TYPE_UNFOCUS:
+			seat_focus_surface(&server->seat, NULL);
+			break;
 		case ACTION_TYPE_ICONIFY:
 			if (view) {
 				view_minimize(view, true);
@@ -665,27 +780,32 @@ actions_run(struct view *activator, struct server *server,
 			break;
 		case ACTION_TYPE_RESIZE_RELATIVE:
 			if (view) {
-				int left = get_arg_value_int(action, "left", 0);
-				int right = get_arg_value_int(action, "right", 0);
-				int top = get_arg_value_int(action, "top", 0);
-				int bottom = get_arg_value_int(action, "bottom", 0);
+				int left = action_get_int(action, "left", 0);
+				int right = action_get_int(action, "right", 0);
+				int top = action_get_int(action, "top", 0);
+				int bottom = action_get_int(action, "bottom", 0);
 				view_resize_relative(view, left, right, top, bottom);
 			}
 			break;
 		case ACTION_TYPE_MOVETO:
 			if (view) {
-				int x = get_arg_value_int(action, "x", 0);
-				int y = get_arg_value_int(action, "y", 0);
+				int x = action_get_int(action, "x", 0);
+				int y = action_get_int(action, "y", 0);
 				view_move(view, x, y);
 			}
 			break;
 		case ACTION_TYPE_MOVE_RELATIVE:
 			if (view) {
-				int x = get_arg_value_int(action, "x", 0);
-				int y = get_arg_value_int(action, "y", 0);
+				int x = action_get_int(action, "x", 0);
+				int y = action_get_int(action, "y", 0);
 				view_move_relative(view, x, y);
 			}
 			break;
+		case ACTION_TYPE_MOVETO_CURSOR:
+			if (view) {
+				view_move_to_cursor(view);
+			}
+			break;
 		case ACTION_TYPE_SEND_TO_DESKTOP:
 			if (!view) {
 				break;
@@ -694,8 +814,8 @@ actions_run(struct view *activator, struct server *server,
 		case ACTION_TYPE_GO_TO_DESKTOP:
 			{
 				bool follow = true;
-				bool wrap = get_arg_value_bool(action, "wrap", true);
-				const char *to = get_arg_value_str(action, "to", NULL);
+				bool wrap = action_get_bool(action, "wrap", true);
+				const char *to = action_get_str(action, "to", NULL);
 				/*
 				 * `to` is always != NULL here because otherwise we would have
 				 * removed the action during the initial parsing step as it is
@@ -708,10 +828,11 @@ actions_run(struct view *activator, struct server *server,
 				}
 				if (action->type == ACTION_TYPE_SEND_TO_DESKTOP) {
 					view_move_to_workspace(view, target);
-					follow = get_arg_value_bool(action, "follow", true);
+					follow = action_get_bool(action, "follow", true);
 				}
 				if (follow) {
-					workspaces_switch_to(target);
+					workspaces_switch_to(target,
+						/*update_focus*/ true);
 				}
 			}
 			break;
@@ -723,7 +844,7 @@ actions_run(struct view *activator, struct server *server,
 			if (!output) {
 				break;
 			}
-			const char *region_name = action_str_from_arg(arg);
+			const char *region_name = action_get_str(action, "region", NULL);
 			struct region *region = regions_from_name(region_name, output);
 			if (region) {
 				view_snap_to_region(view, region,
@@ -739,10 +860,27 @@ actions_run(struct view *activator, struct server *server,
 			break;
 		case ACTION_TYPE_FOCUS_OUTPUT:
 			{
-				const char *output_name = action_str_from_arg(arg);
+				const char *output_name = action_get_str(action, "output", NULL);
 				desktop_focus_output(output_from_name(server, output_name));
 			}
 			break;
+		case ACTION_TYPE_IF:
+			if (view) {
+				run_if_action(view, server, action);
+			}
+			break;
+		case ACTION_TYPE_FOR_EACH:
+			{
+				struct wl_array views;
+				struct view **item;
+				wl_array_init(&views);
+				view_array_append(server, &views, LAB_VIEW_CRITERIA_NONE);
+				wl_array_for_each(item, &views) {
+					run_if_action(*item, server, action);
+				}
+				wl_array_release(&views);
+			}
+			break;
 		case ACTION_TYPE_INVALID:
 			wlr_log(WLR_ERROR, "Not executing unknown action");
 			break;
