@@ -17,6 +17,7 @@
 #include <wlr/util/log.h>
 #include "action.h"
 #include "common/list.h"
+#include "common/macros.h"
 #include "common/mem.h"
 #include "common/nodename.h"
 #include "common/parse-bool.h"
@@ -27,6 +28,7 @@
 #include "config/rcxml.h"
 #include "labwc.h"
 #include "regions.h"
+#include "view.h"
 #include "window-rules.h"
 #include "workspaces.h"
 
@@ -37,6 +39,9 @@ static bool in_mousebind;
 static bool in_libinput_category;
 static bool in_window_switcher_field;
 static bool in_window_rules;
+static bool in_action_query;
+static bool in_action_then_branch;
+static bool in_action_else_branch;
 
 static struct usable_area_override *current_usable_area_override;
 static struct keybind *current_keybind;
@@ -49,6 +54,8 @@ static struct region *current_region;
 static struct window_switcher_field *current_field;
 static struct window_rule *current_window_rule;
 static struct action *current_window_rule_action;
+static struct view_query *current_view_query;
+static struct action *current_child_action;
 
 enum font_place {
 	FONT_PLACE_NONE = 0,
@@ -152,6 +159,8 @@ fill_window_rule(char *nodename, char *content)
 		set_property(content, &current_window_rule->skip_window_switcher);
 	} else if (!strcasecmp(nodename, "ignoreFocusRequest")) {
 		set_property(content, &current_window_rule->ignore_focus_request);
+	} else if (!strcasecmp(nodename, "fixedPosition")) {
+		set_property(content, &current_window_rule->fixed_position);
 
 	/* Actions */
 	} else if (!strcmp(nodename, "name.action")) {
@@ -246,6 +255,79 @@ fill_region(char *nodename, char *content)
 	}
 }
 
+static void
+fill_action_query(char *nodename, char *content, struct action *action)
+{
+	string_truncate_at_pattern(nodename, ".keybind.keyboard");
+	string_truncate_at_pattern(nodename, ".mousebind.context.mouse");
+
+	if (!strcasecmp(nodename, "query.action")) {
+		current_view_query = NULL;
+	}
+
+	string_truncate_at_pattern(nodename, ".query.action");
+
+	if (!content) {
+		return;
+	}
+
+	if (!current_view_query) {
+		struct wl_list *queries = action_get_querylist(action, "query");
+		if (!queries) {
+			action_arg_add_querylist(action, "query");
+			queries = action_get_querylist(action, "query");
+		}
+		current_view_query = znew(*current_view_query);
+		wl_list_append(queries, &current_view_query->link);
+	}
+
+	if (!strcasecmp(nodename, "identifier")) {
+		current_view_query->identifier = xstrdup(content);
+	} else if (!strcasecmp(nodename, "title")) {
+		current_view_query->title = xstrdup(content);
+	}
+}
+
+static void
+fill_child_action(char *nodename, char *content, struct action *parent,
+	const char *branch_name)
+{
+	string_truncate_at_pattern(nodename, ".keybind.keyboard");
+	string_truncate_at_pattern(nodename, ".mousebind.context.mouse");
+	string_truncate_at_pattern(nodename, ".then.action");
+	string_truncate_at_pattern(nodename, ".else.action");
+
+	if (!strcasecmp(nodename, "action")) {
+		current_child_action = NULL;
+	}
+
+	if (!content) {
+		return;
+	}
+
+	struct wl_list *siblings = action_get_actionlist(parent, branch_name);
+	if (!siblings) {
+		action_arg_add_actionlist(parent, branch_name);
+		siblings = action_get_actionlist(parent, branch_name);
+	}
+
+	if (!strcasecmp(nodename, "name.action")) {
+		if (!strcasecmp(content, "If") || !strcasecmp(content, "ForEach")) {
+			wlr_log(WLR_ERROR, "action '%s' cannot be a child action", content);
+			return;
+		}
+		current_child_action = action_create(content);
+		if (current_child_action) {
+			wl_list_append(siblings, &current_child_action->link);
+		}
+	} else if (!current_child_action) {
+		wlr_log(WLR_ERROR, "expect <action name=\"\"> element first. "
+			"nodename: '%s' content: '%s'", nodename, content);
+	} else {
+		action_arg_from_xml_node(current_child_action, nodename, content);
+	}
+}
+
 static void
 fill_keybind(char *nodename, char *content)
 {
@@ -539,10 +621,32 @@ entry(xmlNode *node, char *nodename, char *content)
 		fill_usable_area_override(nodename, content);
 	}
 	if (in_keybind) {
-		fill_keybind(nodename, content);
+		if (in_action_query) {
+			fill_action_query(nodename, content,
+				current_keybind_action);
+		} else if (in_action_then_branch) {
+			fill_child_action(nodename, content,
+				current_keybind_action, "then");
+		} else if (in_action_else_branch) {
+			fill_child_action(nodename, content,
+				current_keybind_action, "else");
+		} else {
+			fill_keybind(nodename, content);
+		}
 	}
 	if (in_mousebind) {
-		fill_mousebind(nodename, content);
+		if (in_action_query) {
+			fill_action_query(nodename, content,
+				current_mousebind_action);
+		} else if (in_action_then_branch) {
+			fill_child_action(nodename, content,
+				current_mousebind_action, "then");
+		} else if (in_action_else_branch) {
+			fill_child_action(nodename, content,
+				current_mousebind_action, "else");
+		} else {
+			fill_mousebind(nodename, content);
+		}
 	}
 	if (in_libinput_category) {
 		fill_libinput_category(nodename, content);
@@ -631,6 +735,14 @@ entry(xmlNode *node, char *nodename, char *content)
 		rc.repeat_rate = atoi(content);
 	} else if (!strcasecmp(nodename, "repeatDelay.keyboard")) {
 		rc.repeat_delay = atoi(content);
+	} else if (!strcasecmp(nodename, "numlock.keyboard")) {
+		set_bool(content, &rc.kb_numlock_enable);
+	} else if (!strcasecmp(nodename, "layoutScope.keyboard")) {
+		/*
+		 * This can be changed to an enum later on
+		 * if we decide to also support "application".
+		 */
+		rc.kb_layout_per_window = !strcasecmp(content, "window");
 	} else if (!strcasecmp(nodename, "screenEdgeStrength.resistance")) {
 		rc.screen_edge_strength = atoi(content);
 	} else if (!strcasecmp(nodename, "range.snapping")) {
@@ -771,6 +883,24 @@ xml_tree_walk(xmlNode *node)
 			in_window_rules = false;
 			continue;
 		}
+		if (!strcasecmp((char *)n->name, "query")) {
+			in_action_query = true;
+			traverse(n);
+			in_action_query = false;
+			continue;
+		}
+		if (!strcasecmp((char *)n->name, "then")) {
+			in_action_then_branch = true;
+			traverse(n);
+			in_action_then_branch = false;
+			continue;
+		}
+		if (!strcasecmp((char *)n->name, "else")) {
+			in_action_else_branch = true;
+			traverse(n);
+			in_action_else_branch = false;
+			continue;
+		}
 		traverse(n);
 	}
 }
@@ -830,6 +960,8 @@ rcxml_init(void)
 	rc.scroll_factor = 1.0;
 	rc.repeat_rate = 25;
 	rc.repeat_delay = 600;
+	rc.kb_numlock_enable = true;
+	rc.kb_layout_per_window = false;
 	rc.screen_edge_strength = 20;
 
 	rc.snap_edge_range = 1;
@@ -941,11 +1073,14 @@ static struct mouse_combos {
 	{ "Title", "Left", "DoubleClick", "ToggleMaximize", NULL, NULL },
 	{ "TitleBar", "Right", "Click", "Focus", NULL, NULL},
 	{ "TitleBar", "Right", "Click", "Raise", NULL, NULL},
-	{ "TitleBar", "Right", "Click", "ShowMenu", "menu", "client-menu"},
+	{ "Title", "Right", "Click", "ShowMenu", "menu", "client-menu"},
 	{ "Close", "Left", "Click", "Close", NULL, NULL },
 	{ "Iconify", "Left", "Click", "Iconify", NULL, NULL},
 	{ "Maximize", "Left", "Click", "ToggleMaximize", NULL, NULL},
+	{ "Maximize", "Right", "Click", "ToggleMaximize", "direction", "horizontal"},
+	{ "Maximize", "Middle", "Click", "ToggleMaximize", "direction", "vertical"},
 	{ "WindowMenu", "Left", "Click", "ShowMenu", "menu", "client-menu"},
+	{ "WindowMenu", "Right", "Click", "ShowMenu", "menu", "client-menu"},
 	{ "Root", "Left", "Press", "ShowMenu", "menu", "root-menu"},
 	{ "Root", "Right", "Press", "ShowMenu", "menu", "root-menu"},
 	{ "Root", "Middle", "Press", "ShowMenu", "menu", "root-menu"},
@@ -1366,6 +1501,8 @@ rcxml_finish(void)
 	current_mouse_context = NULL;
 	current_keybind_action = NULL;
 	current_mousebind_action = NULL;
+	current_child_action = NULL;
+	current_view_query = NULL;
 	current_region = NULL;
 	current_field = NULL;
 	current_window_rule = NULL;
