@@ -13,6 +13,10 @@
 #include "workspaces.h"
 #include "xwayland.h"
 
+#if HAVE_XWAYLAND
+#include <wlr/xwayland.h>
+#endif
+
 void
 desktop_arrange_all_views(struct server *server)
 {
@@ -35,31 +39,21 @@ desktop_arrange_all_views(struct server *server)
 }
 
 void
-desktop_focus_and_activate_view(struct seat *seat, struct view *view)
+desktop_focus_view(struct view *view, bool raise)
 {
-	if (!view) {
-		seat_focus_surface(seat, NULL);
-		return;
-	}
-
+	assert(view);
 	/*
 	 * Guard against views with no mapped surfaces when handling
 	 * 'request_activate' and 'request_minimize'.
-	 * See notes by view_isfocusable()
 	 */
 	if (!view->surface) {
 		return;
 	}
 
-	if (input_inhibit_blocks_surface(seat, view->surface->resource)
-			|| seat->server->session_lock) {
-		return;
-	}
-
 	if (view->minimized) {
 		/*
 		 * Unminimizing will map the view which triggers a call to this
-		 * function again.
+		 * function again (with raise=true).
 		 */
 		view_minimize(view, false);
 		return;
@@ -69,16 +63,48 @@ desktop_focus_and_activate_view(struct seat *seat, struct view *view)
 		return;
 	}
 
-	struct wlr_surface *prev_surface;
-	prev_surface = seat->seat->keyboard_state.focused_surface;
-
-	/* Do not re-focus an already focused surface. */
-	if (prev_surface == view->surface) {
-		return;
+	/*
+	 * Switch workspace if necessary to make the view visible
+	 * (unnecessary for "always on {top,bottom}" views).
+	 */
+	if (!view_is_always_on_top(view) && !view_is_always_on_bottom(view)) {
+		workspaces_switch_to(view->workspace, /*update_focus*/ false);
 	}
 
-	view_set_activated(view);
-	seat_focus_surface(seat, view->surface);
+	/*
+	 * Give input focus, even if the view claims not to want it (see
+	 * view->impl->wants_focus). This is a workaround for so-called
+	 * "globally active" X11 views (MATLAB known to be one such)
+	 * that expect to be able to control focus themselves, but can't
+	 * under labwc since it's disallowed at the wlroots level.
+	 */
+	struct seat *seat = &view->server->seat;
+	if (view->surface != seat->seat->keyboard_state.focused_surface) {
+		seat_focus_surface(seat, view->surface);
+	}
+
+	if (raise) {
+		view_move_to_front(view);
+	}
+}
+
+/* TODO: focus layer-shell surfaces also? */
+void
+desktop_focus_view_or_surface(struct seat *seat, struct view *view,
+		struct wlr_surface *surface, bool raise)
+{
+	assert(view || surface);
+	if (view) {
+		desktop_focus_view(view, raise);
+#if HAVE_XWAYLAND
+	} else if (wlr_surface_is_xwayland_surface(surface)) {
+		struct wlr_xwayland_surface *xsurface =
+			wlr_xwayland_surface_from_wlr_surface(surface);
+		if (xsurface && wlr_xwayland_or_surface_wants_focus(xsurface)) {
+			seat_focus_surface(seat, surface);
+		}
+#endif
+	}
 }
 
 static struct wl_list *
@@ -105,7 +131,7 @@ first_view(struct server *server)
 			continue;
 		}
 		struct view *view = node_view_from_node(node);
-		if (view_isfocusable(view)) {
+		if (view_is_focusable(view)) {
 			return view;
 		}
 	}
@@ -147,7 +173,7 @@ desktop_cycle_view(struct server *server, struct view *start_view,
 
 	if (!start_view) {
 		start_view = first_view(server);
-		if (!start_view || start_view != desktop_focused_view(server)) {
+		if (!start_view || start_view != server->focused_view) {
 			return start_view;  /* may be NULL */
 		}
 	}
@@ -177,7 +203,7 @@ desktop_cycle_view(struct server *server, struct view *start_view,
 		view = node_view_from_node(node);
 
 		enum property skip = window_rules_get_property(view, "skipWindowSwitcher");
-		if (view_isfocusable(view) && skip != LAB_PROP_TRUE) {
+		if (view_is_focusable(view) && skip != LAB_PROP_TRUE) {
 			return view;
 		}
 	} while (view != start_view);
@@ -187,8 +213,10 @@ desktop_cycle_view(struct server *server, struct view *start_view,
 }
 
 struct view *
-desktop_topmost_mapped_view(struct server *server)
+desktop_topmost_focusable_view(struct server *server)
 {
+	struct wlr_surface *prev =
+		server->seat.seat->keyboard_state.focused_surface;
 	struct view *view;
 	struct wl_list *node_list;
 	struct wlr_scene_node *node;
@@ -199,39 +227,25 @@ desktop_topmost_mapped_view(struct server *server)
 			continue;
 		}
 		view = node_view_from_node(node);
-		if (view->mapped) {
+		if (view->mapped && view_is_focusable_from(view, prev)) {
 			return view;
 		}
 	}
 	return NULL;
 }
 
-struct view *
-desktop_focused_view(struct server *server)
-{
-	struct seat *seat = &server->seat;
-	struct wlr_surface *focused_surface;
-	focused_surface = seat->seat->keyboard_state.focused_surface;
-	if (!focused_surface) {
-		return NULL;
-	}
-	struct view *view;
-	wl_list_for_each(view, &server->views, link) {
-		if (view->surface == focused_surface) {
-			return view;
-		}
-	}
-
-	return NULL;
-}
-
 void
-desktop_focus_topmost_mapped_view(struct server *server)
+desktop_focus_topmost_view(struct server *server)
 {
-	struct view *view = desktop_topmost_mapped_view(server);
-	desktop_focus_and_activate_view(&server->seat, view);
+	struct view *view = desktop_topmost_focusable_view(server);
 	if (view) {
-		view_move_to_front(view);
+		desktop_focus_view(view, /*raise*/ true);
+	} else {
+		/*
+		 * Defocus previous focused surface/view if no longer
+		 * focusable (e.g. unmapped or on a different workspace).
+		 */
+		seat_focus_surface(&server->seat, NULL);
 	}
 }
 
@@ -252,12 +266,12 @@ desktop_focus_output(struct output *output)
 			continue;
 		}
 		view = node_view_from_node(node);
-		if (!view_isfocusable(view)) {
+		if (!view_is_focusable(view)) {
 			continue;
 		}
 		if (wlr_output_layout_intersects(layout,
 				output->wlr_output, &view->current)) {
-			desktop_focus_and_activate_view(&output->server->seat, view);
+			desktop_focus_view(view, /*raise*/ false);
 			wlr_cursor_warp(output->server->seat.cursor, NULL,
 				view->current.x + view->current.width / 2,
 				view->current.y + view->current.height / 2);
@@ -275,6 +289,32 @@ desktop_focus_output(struct output *output)
 	cursor_update_focus(output->server);
 }
 
+void
+desktop_update_top_layer_visiblity(struct server *server)
+{
+	struct view *view;
+	struct output *output;
+	uint32_t top = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
+
+	/* Enable all top layers */
+	wl_list_for_each(output, &server->outputs, link) {
+		if (!output_is_usable(output)) {
+			continue;
+		}
+		wlr_scene_node_set_enabled(&output->layer_tree[top]->node, true);
+	}
+
+	/* And disable them again when there is a view in fullscreen */
+	enum lab_view_criteria criteria =
+		LAB_VIEW_CRITERIA_CURRENT_WORKSPACE | LAB_VIEW_CRITERIA_FULLSCREEN;
+	for_each_view(view, &server->views, criteria) {
+		if (!output_is_usable(view->output)) {
+			continue;
+		}
+		wlr_scene_node_set_enabled(&view->output->layer_tree[top]->node, false);
+	}
+}
+
 static struct wlr_surface *
 get_surface_from_layer_node(struct wlr_scene_node *node)
 {
