@@ -1,8 +1,10 @@
 // SPDX-License-Identifier: GPL-2.0-only
+#include "input/keyboard.h"
 #include "labwc.h"
 #include "regions.h"
 #include "resize_indicator.h"
 #include "view.h"
+#include "window-rules.h"
 
 static int
 max_move_scale(double pos_cursor, double pos_current,
@@ -33,6 +35,10 @@ interactive_begin(struct view *view, enum input_mode mode, uint32_t edges)
 		return;
 	}
 
+	if (window_rules_get_property(view, "fixedPosition") == LAB_PROP_TRUE) {
+		return;
+	}
+
 	switch (mode) {
 	case LAB_INPUT_STATE_MOVE:
 		if (view->fullscreen) {
@@ -45,7 +51,7 @@ interactive_begin(struct view *view, enum input_mode mode, uint32_t edges)
 			 */
 			return;
 		}
-		if (view->maximized || view_is_tiled(view)) {
+		if (!view_is_floating(view)) {
 			/*
 			 * Un-maximize and restore natural width/height.
 			 * Don't reset tiled state yet since we may want
@@ -71,18 +77,20 @@ interactive_begin(struct view *view, enum input_mode mode, uint32_t edges)
 		cursor_set(seat, LAB_CURSOR_GRAB);
 		break;
 	case LAB_INPUT_STATE_RESIZE:
-		if (view->maximized || view->fullscreen) {
+		if (view->fullscreen || view->maximized == VIEW_AXIS_BOTH) {
 			/*
-			 * We don't allow resizing while maximized or
-			 * fullscreen.
+			 * We don't allow resizing while fullscreen or
+			 * maximized in both directions.
 			 */
 			return;
 		}
 		/*
-		 * Reset tiled state but keep the same geometry as the
-		 * starting point for the resize.
+		 * If tiled or maximized in only one direction, reset
+		 * tiled/maximized state but keep the same geometry as
+		 * the starting point for the resize.
 		 */
 		view_set_untiled(view);
+		view_restore_to(view, view->pending);
 		cursor_set(seat, cursor_get_from_edge(edges));
 		break;
 	default:
@@ -111,37 +119,45 @@ snap_to_edge(struct view *view)
 		return false;
 	}
 
+	struct output *output = output_nearest_to_cursor(view->server);
+	if (!output_is_usable(output)) {
+		wlr_log(WLR_ERROR, "output at cursor is unusable");
+		return false;
+	}
+
 	/* Translate into output local coordinates */
 	double cursor_x = view->server->seat.cursor->x;
 	double cursor_y = view->server->seat.cursor->y;
 	wlr_output_layout_output_coords(view->server->output_layout,
-		view->output->wlr_output, &cursor_x, &cursor_y);
+		output->wlr_output, &cursor_x, &cursor_y);
 
+	struct wlr_box *area = &output->usable_area;
+	enum view_edge edge;
+	if (cursor_x <= area->x + snap_range) {
+		edge = VIEW_EDGE_LEFT;
+	} else if (cursor_x >= area->x + area->width - snap_range) {
+		edge = VIEW_EDGE_RIGHT;
+	} else if (cursor_y <= area->y + snap_range) {
+		edge = VIEW_EDGE_UP;
+	} else if (cursor_y >= area->y + area->height - snap_range) {
+		edge = VIEW_EDGE_DOWN;
+	} else {
+		/* Not close to any edge */
+		return false;
+	}
+
+	view_set_output(view, output);
 	/*
 	 * Don't store natural geometry here (it was
 	 * stored already in interactive_begin())
 	 */
-	struct wlr_box *area = &view->output->usable_area;
-	if (cursor_x <= area->x + snap_range) {
-		view_snap_to_edge(view, VIEW_EDGE_LEFT,
-			/*store_natural_geometry*/ false);
-	} else if (cursor_x >= area->x + area->width - snap_range) {
-		view_snap_to_edge(view, VIEW_EDGE_RIGHT,
-			/*store_natural_geometry*/ false);
-	} else if (cursor_y <= area->y + snap_range) {
-		if (rc.snap_top_maximize) {
-			view_maximize(view, true,
-				/*store_natural_geometry*/ false);
-		} else {
-			view_snap_to_edge(view, VIEW_EDGE_UP,
-				/*store_natural_geometry*/ false);
-		}
-	} else if (cursor_y >= area->y + area->height - snap_range) {
-		view_snap_to_edge(view, VIEW_EDGE_DOWN,
+	if (edge == VIEW_EDGE_UP && rc.snap_top_maximize) {
+		view_maximize(view, VIEW_AXIS_BOTH,
 			/*store_natural_geometry*/ false);
 	} else {
-		/* Not close to any edge */
-		return false;
+		view_snap_to_edge(view, edge,
+			/*across_outputs*/ false,
+			/*store_natural_geometry*/ false);
 	}
 
 	return true;
