@@ -33,22 +33,14 @@ static struct action *current_item_action;
 static int menu_level;
 static struct menu *current_menu;
 
-/* vector for <menu id="" label=""> elements */
-static struct menu *menus;
-static int nr_menus, alloc_menus;
-
 /* TODO: split this whole file into parser.c and actions.c*/
 
 static struct menu *
 menu_create(struct server *server, const char *id, const char *label)
 {
-	if (nr_menus == alloc_menus) {
-		alloc_menus = (alloc_menus + 16) * 2;
-		menus = xrealloc(menus, alloc_menus * sizeof(struct menu));
-	}
-	struct menu *menu = menus + nr_menus;
-	memset(menu, 0, sizeof(*menu));
-	nr_menus++;
+	struct menu *menu = znew(*menu);
+	wl_list_append(&server->menus, &menu->link);
+
 	wl_list_init(&menu->menuitems);
 	menu->id = xstrdup(id);
 	menu->label = xstrdup(label ? label : id);
@@ -62,14 +54,13 @@ menu_create(struct server *server, const char *id, const char *label)
 }
 
 struct menu *
-menu_get_by_id(const char *id)
+menu_get_by_id(struct server *server, const char *id)
 {
 	if (!id) {
 		return NULL;
 	}
 	struct menu *menu;
-	for (int i = 0; i < nr_menus; ++i) {
-		menu = menus + i;
+	wl_list_for_each(menu, &server->menus, link) {
 		if (!strcmp(menu->id, id)) {
 			return menu;
 		}
@@ -124,8 +115,7 @@ static void
 post_processing(struct server *server)
 {
 	struct menu *menu;
-	for (int i = 0; i < nr_menus; ++i) {
-		menu = menus + i;
+	wl_list_for_each(menu, &server->menus, link) {
 		menu_update_width(menu);
 	}
 }
@@ -149,8 +139,9 @@ validate_menu(struct menu *menu)
 static void
 validate(struct server *server)
 {
-	for (int i = 0; i < nr_menus; ++i) {
-		validate_menu(menus + i);
+	struct menu *menu;
+	wl_list_for_each(menu, &server->menus, link) {
+		validate_menu(menu);
 	}
 }
 
@@ -201,7 +192,7 @@ item_create(struct menu *menu, const char *text, bool show_arrow)
 	menuitem->selected.buffer = scaled_font_buffer_create(menuitem->selected.tree);
 	if (!menuitem->normal.buffer || !menuitem->selected.buffer) {
 		wlr_log(WLR_ERROR, "Failed to create menu item '%s'", text);
-		/**
+		/*
 		 * Destroying the root node will destroy everything,
 		 * including the node descriptor and scaled_font_buffers.
 		 */
@@ -473,7 +464,7 @@ handle_menu_element(xmlNode *n, struct server *server)
 		 * <menu id=""> creates an entry which points to a menu
 		 * defined elsewhere
 		 */
-		struct menu *menu = menu_get_by_id(id);
+		struct menu *menu = menu_get_by_id(server, id);
 		if (menu) {
 			current_item = item_create(current_menu, menu->label, true);
 			if (current_item) {
@@ -483,9 +474,9 @@ handle_menu_element(xmlNode *n, struct server *server)
 			wlr_log(WLR_ERROR, "no menu with id '%s'", id);
 		}
 	}
-	zfree(label);
-	zfree(execute);
-	zfree(id);
+	free(label);
+	free(execute);
+	free(id);
 }
 
 /* This can be one of <separator> and <separator label=""> */
@@ -494,9 +485,7 @@ handle_separator_element(xmlNode *n)
 {
 	char *label = (char *)xmlGetProp(n, (const xmlChar *)"label");
 	current_item = separator_create(current_menu, label);
-	if (label) {
-		free(label);
-	}
+	free(label);
 }
 
 static void
@@ -524,26 +513,18 @@ xml_tree_walk(xmlNode *node, struct server *server)
 	}
 }
 
+/*
+ * @stream can come from either of the following:
+ *   - fopen() in the case of reading a file such as menu.xml
+ *   - popen() when processing pipemenus
+ */
 static void
-parse_xml(const char *filename, struct server *server)
+parse(struct server *server, FILE *stream)
 {
-	FILE *stream;
 	char *line = NULL;
 	size_t len = 0;
 	struct buf b;
-	static char menuxml[4096] = { 0 };
 
-	if (!rc.config_dir) {
-		return;
-	}
-	snprintf(menuxml, sizeof(menuxml), "%s/%s", rc.config_dir, filename);
-
-	stream = fopen(menuxml, "r");
-	if (!stream) {
-		wlr_log(WLR_ERROR, "cannot read %s", menuxml);
-		return;
-	}
-	wlr_log(WLR_INFO, "read menu file %s", menuxml);
 	buf_init(&b);
 	while (getline(&line, &len, stream) != -1) {
 		char *p = strrchr(line, '\n');
@@ -553,7 +534,6 @@ parse_xml(const char *filename, struct server *server)
 		buf_add(&b, line);
 	}
 	free(line);
-	fclose(stream);
 	xmlDoc *d = xmlParseMemory(b.buf, b.len);
 	if (!d) {
 		wlr_log(WLR_ERROR, "xmlParseMemory()");
@@ -566,6 +546,26 @@ err:
 	free(b.buf);
 }
 
+static void
+parse_xml(const char *filename, struct server *server)
+{
+	static char buf[4096] = { 0 };
+
+	if (!rc.config_dir) {
+		return;
+	}
+	snprintf(buf, sizeof(buf), "%s/%s", rc.config_dir, filename);
+
+	FILE *stream = fopen(buf, "r");
+	if (!stream) {
+		wlr_log(WLR_ERROR, "cannot read %s", buf);
+		return;
+	}
+	wlr_log(WLR_INFO, "read menu file %s", buf);
+	parse(server, stream);
+	fclose(stream);
+}
+
 static int
 menu_get_full_width(struct menu *menu)
 {
@@ -654,15 +654,14 @@ menu_configure(struct menu *menu, int lx, int ly, enum menu_align align)
 }
 
 static void
-menu_hide_submenu(const char *id)
+menu_hide_submenu(struct server *server, const char *id)
 {
 	struct menu *menu, *hide_menu;
-	hide_menu = menu_get_by_id(id);
+	hide_menu = menu_get_by_id(server, id);
 	if (!hide_menu) {
 		return;
 	}
-	for (int i = 0; i < nr_menus; ++i) {
-		menu = menus + i;
+	wl_list_for_each(menu, &server->menus, link) {
 		bool should_reposition = false;
 		struct menuitem *item, *next;
 		wl_list_for_each_safe(item, next, &menu->menuitems, link) {
@@ -688,7 +687,7 @@ menu_hide_submenu(const char *id)
 static void
 init_rootmenu(struct server *server)
 {
-	struct menu *menu = menu_get_by_id("root-menu");
+	struct menu *menu = menu_get_by_id(server, "root-menu");
 
 	/* Default menu if no menu.xml found */
 	if (!menu) {
@@ -706,7 +705,7 @@ init_rootmenu(struct server *server)
 static void
 init_windowmenu(struct server *server)
 {
-	struct menu *menu = menu_get_by_id("client-menu");
+	struct menu *menu = menu_get_by_id(server, "client-menu");
 
 	/* Default menu if no menu.xml found */
 	if (!menu) {
@@ -745,13 +744,14 @@ init_windowmenu(struct server *server)
 	}
 
 	if (wl_list_length(&rc.workspace_config.workspaces) == 1) {
-		menu_hide_submenu("workspaces");
+		menu_hide_submenu(server, "workspaces");
 	}
 }
 
 void
 menu_init(struct server *server)
 {
+	wl_list_init(&server->menus);
 	parse_xml("menu.xml", server);
 	init_rootmenu(server);
 	init_windowmenu(server);
@@ -760,11 +760,10 @@ menu_init(struct server *server)
 }
 
 void
-menu_finish(void)
+menu_finish(struct server *server)
 {
-	struct menu *menu;
-	for (int i = 0; i < nr_menus; ++i) {
-		menu = menus + i;
+	struct menu *menu, *tmp_menu;
+	wl_list_for_each_safe(menu, tmp_menu, &server->menus, link) {
 		struct menuitem *item, *next;
 		wl_list_for_each_safe(item, next, &menu->menuitems, link) {
 			item_destroy(item);
@@ -774,10 +773,9 @@ menu_finish(void)
 		 * including node descriptors and scaled_font_buffers.
 		 */
 		wlr_scene_node_destroy(&menu->scene_tree->node);
+		wl_list_remove(&menu->link);
+		zfree(menu);
 	}
-	zfree(menus);
-	alloc_menus = 0;
-	nr_menus = 0;
 }
 
 /* Sets selection (or clears selection if passing NULL) */
@@ -932,7 +930,7 @@ menu_execute_item(struct menuitem *item)
 	actions_run(item->parent->triggered_by_view,
 		item->parent->server, &item->actions, 0);
 
-	/**
+	/*
 	 * We close the menu here to provide a faster feedback to the user.
 	 * We do that without resetting the input state so src/cursor.c
 	 * can do its own clean up on the following RELEASE event.
@@ -1040,7 +1038,7 @@ menu_close_root(struct server *server)
 void
 menu_reconfigure(struct server *server)
 {
-	menu_finish();
+	menu_finish(server);
 	server->menu_current = NULL;
 	menu_init(server);
 }
