@@ -16,7 +16,7 @@
 #include <wlr/types/wlr_scene.h>
 #include <wlr/util/region.h>
 #include <wlr/util/log.h>
-#include "common/array-size.h"
+#include "common/macros.h"
 #include "common/mem.h"
 #include "labwc.h"
 #include "layers.h"
@@ -272,6 +272,7 @@ static void
 output_update_for_layout_change(struct server *server)
 {
 	output_update_all_usable_areas(server, /*layout_changed*/ true);
+	session_lock_update_for_layout_change();
 
 	/*
 	 * "Move" each wlr_output_cursor (in per-output coordinates) to
@@ -298,7 +299,7 @@ output_config_apply(struct server *server,
 
 		wlr_output_enable(o, output_enabled);
 		if (output_enabled) {
-			/* Output specifc actions only */
+			/* Output specific actions only */
 			if (head->state.mode) {
 				wlr_output_set_mode(o, head->state.mode);
 			} else {
@@ -564,9 +565,29 @@ output_usable_area_in_layout_coords(struct output *output)
 	return box;
 }
 
+struct wlr_box
+output_usable_area_scaled(struct output *output)
+{
+	if (!output) {
+		return (struct wlr_box){0};
+	}
+	struct wlr_box usable = output_usable_area_in_layout_coords(output);
+	if (usable.height == output->wlr_output->height
+			&& output->wlr_output->scale != 1) {
+		usable.height /= output->wlr_output->scale;
+	}
+	if (usable.width == output->wlr_output->width
+			&& output->wlr_output->scale != 1) {
+		usable.width /= output->wlr_output->scale;
+	}
+	return usable;
+}
+
 void
 handle_output_power_manager_set_mode(struct wl_listener *listener, void *data)
 {
+	struct server *server = wl_container_of(listener, server,
+		output_power_manager_set_mode);
 	struct wlr_output_power_v1_set_mode_event *event = data;
 
 	switch (event->mode) {
@@ -580,6 +601,11 @@ handle_output_power_manager_set_mode(struct wl_listener *listener, void *data)
 			wlr_output_rollback(event->output);
 		}
 		wlr_output_commit(event->output);
+		/*
+		 * Re-set the cursor image so that the cursor
+		 * isn't invisible on the newly enabled output.
+		 */
+		cursor_update_image(&server->seat);
 		break;
 	}
 }
