@@ -9,8 +9,11 @@
 #include <wlr/types/wlr_touch.h>
 #include <wlr/util/log.h>
 #include "common/mem.h"
-#include "key-state.h"
+#include "input/input.h"
+#include "input/keyboard.h"
+#include "input/key-state.h"
 #include "labwc.h"
+#include "view.h"
 
 static void
 input_device_destroy(struct wl_listener *listener, void *data)
@@ -210,14 +213,19 @@ new_keyboard(struct seat *seat, struct wlr_input_device *device, bool virtual)
 
 	wlr_keyboard_set_keymap(kb, seat->keyboard_group->keyboard.keymap);
 
+	/*
+	 * This needs to be before wlr_keyboard_group_add_keyboard().
+	 * For some reason, wlroots takes the modifier state from the
+	 * new keyboard and syncs it to the others in the group, rather
+	 * than the other way around.
+	 */
+	keyboard_set_numlock(kb);
+
 	if (!virtual) {
 		wlr_keyboard_group_add_keyboard(seat->keyboard_group, kb);
 	}
 
-	keyboard->key.notify = keyboard_key_notify;
-	wl_signal_add(&kb->events.key, &keyboard->key);
-	keyboard->modifier.notify = keyboard_modifiers_notify;
-	wl_signal_add(&kb->events.modifiers, &keyboard->modifier);
+	keyboard_setup_handlers(keyboard);
 
 	wlr_seat_set_keyboard(seat->seat, kb);
 
@@ -335,6 +343,38 @@ new_virtual_keyboard(struct wl_listener *listener, void *data)
 	seat_add_device(seat, input);
 }
 
+static void
+focus_change_notify(struct wl_listener *listener, void *data)
+{
+	struct seat *seat = wl_container_of(listener, seat, focus_change);
+	struct wlr_seat_keyboard_focus_change_event *event = data;
+	struct server *server = seat->server;
+	struct view *view = event->new_surface ?
+		view_from_wlr_surface(event->new_surface) : NULL;
+
+	/*
+	 * If an xwayland-unmanaged surface was focused belonging to the
+	 * same application as the focused view, allow the view to remain
+	 * active. This fixes an issue with menus immediately closing in
+	 * some X11 apps (try LibreOffice with SAL_USE_VCLPLUGIN=gen).
+	 */
+	if (!view && server->focused_view && event->new_surface
+			&& view_is_related(server->focused_view,
+				event->new_surface)) {
+		return;
+	}
+
+	if (view != server->focused_view) {
+		if (server->focused_view) {
+			view_set_activated(server->focused_view, false);
+		}
+		if (view) {
+			view_set_activated(view, true);
+		}
+		server->focused_view = view;
+	}
+}
+
 void
 seat_init(struct server *server)
 {
@@ -353,6 +393,10 @@ seat_init(struct server *server)
 	seat->new_input.notify = new_input_notify;
 	wl_signal_add(&server->backend->events.new_input, &seat->new_input);
 
+	seat->focus_change.notify = focus_change_notify;
+	wl_signal_add(&seat->seat->keyboard_state.events.focus_change,
+		&seat->focus_change);
+
 	seat->virtual_pointer = wlr_virtual_pointer_manager_v1_create(
 		server->wl_display);
 	wl_signal_add(&seat->virtual_pointer->events.new_virtual_pointer,
@@ -372,9 +416,7 @@ seat_init(struct server *server)
 	}
 	wlr_cursor_attach_output_layout(seat->cursor, server->output_layout);
 
-	keyboard_init(seat);
-	cursor_init(seat);
-	touch_init(seat);
+	input_handlers_init(seat);
 }
 
 void
@@ -382,20 +424,14 @@ seat_finish(struct server *server)
 {
 	struct seat *seat = &server->seat;
 	wl_list_remove(&seat->new_input.link);
+	wl_list_remove(&seat->focus_change.link);
 
 	struct input *input, *next;
 	wl_list_for_each_safe(input, next, &seat->inputs, link) {
 		input_device_destroy(&input->destroy, NULL);
 	}
 
-	keyboard_finish(seat);
-	/*
-	 * Caution - touch_finish() unregisters event listeners from
-	 * seat->cursor and must come before cursor_finish(), otherwise
-	 * a use-after-free occurs.
-	 */
-	touch_finish(seat);
-	cursor_finish(seat);
+	input_handlers_finish(seat);
 }
 
 static void
@@ -426,13 +462,27 @@ seat_reconfigure(struct server *server)
 }
 
 static void
-seat_focus(struct seat *seat, struct wlr_surface *surface)
+seat_focus(struct seat *seat, struct wlr_surface *surface, bool is_lock_surface)
 {
+	/*
+	 * Respect session lock. This check is critical, DO NOT REMOVE.
+	 * It should also come before the !surface condition, or the
+	 * lock screen may lose focus and become impossible to unlock.
+	 */
+	struct server *server = seat->server;
+	if (server->session_lock && !is_lock_surface) {
+		return;
+	}
+
 	if (!surface) {
 		wlr_seat_keyboard_notify_clear_focus(seat->seat);
 		return;
 	}
-	struct wlr_keyboard *kb = &seat->keyboard_group->keyboard;
+
+	/* Respect input inhibit (also used by some lock screens) */
+	if (input_inhibit_blocks_surface(seat, surface->resource)) {
+		return;
+	}
 
 	/*
 	 * Key events associated with keybindings (both pressed and released)
@@ -444,10 +494,10 @@ seat_focus(struct seat *seat, struct wlr_surface *surface)
 	uint32_t *pressed_sent_keycodes = key_state_pressed_sent_keycodes();
 	int nr_pressed_sent_keycodes = key_state_nr_pressed_sent_keycodes();
 
+	struct wlr_keyboard *kb = &seat->keyboard_group->keyboard;
 	wlr_seat_keyboard_notify_enter(seat->seat, surface,
 		pressed_sent_keycodes, nr_pressed_sent_keycodes, &kb->modifiers);
 
-	struct server *server = seat->server;
 	struct wlr_pointer_constraint_v1 *constraint =
 		wlr_pointer_constraints_v1_constraint_for_surface(server->constraints,
 			surface, seat->seat);
@@ -457,12 +507,18 @@ seat_focus(struct seat *seat, struct wlr_surface *surface)
 void
 seat_focus_surface(struct seat *seat, struct wlr_surface *surface)
 {
-	/* Respect layer-shell exlusive keyboard-interactivity. */
+	/* Respect layer-shell exclusive keyboard-interactivity. */
 	if (seat->focused_layer && seat->focused_layer->current.keyboard_interactive
 			== ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE) {
 		return;
 	}
-	seat_focus(seat, surface);
+	seat_focus(seat, surface, /*is_lock_surface*/ false);
+}
+
+void
+seat_focus_lock_surface(struct seat *seat, struct wlr_surface *surface)
+{
+	seat_focus(seat, surface, /*is_lock_surface*/ true);
 }
 
 void
@@ -470,10 +526,10 @@ seat_set_focus_layer(struct seat *seat, struct wlr_layer_surface_v1 *layer)
 {
 	if (!layer) {
 		seat->focused_layer = NULL;
-		desktop_focus_topmost_mapped_view(seat->server);
+		desktop_focus_topmost_view(seat->server);
 		return;
 	}
-	seat_focus(seat, layer->surface);
+	seat_focus(seat, layer->surface, /*is_lock_surface*/ false);
 	if (layer->current.layer >= ZWLR_LAYER_SHELL_V1_LAYER_TOP) {
 		seat->focused_layer = layer;
 	}
