@@ -25,34 +25,10 @@ view_impl_move_to_back(struct view *view)
 	wlr_scene_node_lower_to_bottom(&view->scene_tree->node);
 }
 
-void
-view_impl_move_sub_views(struct view *parent, enum z_direction z_direction)
-{
-	assert(parent);
-	if (!parent->impl->append_children) {
-		return;
-	}
-
-	struct wl_array subviews;
-	wl_array_init(&subviews);
-	parent->impl->append_children(parent, &subviews);
-
-	struct view **view;
-	wl_array_for_each(view, &subviews) {
-		if (z_direction == LAB_TO_FRONT) {
-			view_impl_move_to_front(*view);
-		} else if (z_direction == LAB_TO_BACK) {
-			view_impl_move_to_back(*view);
-		}
-	}
-	wl_array_release(&subviews);
-}
-
 void
 view_impl_map(struct view *view)
 {
-	desktop_focus_and_activate_view(&view->server->seat, view);
-	view_move_to_front(view);
+	desktop_focus_view(view, /*raise*/ true);
 	view_update_title(view);
 	view_update_app_id(view);
 	if (!view->been_mapped) {
@@ -76,6 +52,18 @@ view_impl_map(struct view *view)
 		view_get_string_prop(view, "title"));
 }
 
+void
+view_impl_unmap(struct view *view)
+{
+	struct server *server = view->server;
+	if (view == server->focused_view) {
+		desktop_focus_topmost_view(server);
+	}
+	if (view == server->last_raised_view) {
+		server->last_raised_view = NULL;
+	}
+}
+
 static bool
 resizing_edge(struct view *view, uint32_t edge)
 {
