@@ -2,27 +2,91 @@
 #include <assert.h>
 #include <stdio.h>
 #include <strings.h>
+#include "common/macros.h"
+#include "common/match.h"
 #include "common/mem.h"
 #include "common/scene-helpers.h"
+#include "input/keyboard.h"
 #include "labwc.h"
 #include "menu/menu.h"
 #include "regions.h"
 #include "resize_indicator.h"
+#include "snap.h"
 #include "ssd.h"
 #include "view.h"
 #include "window-rules.h"
 #include "workspaces.h"
 #include "xwayland.h"
 
-#define LAB_MIN_VIEW_WIDTH  100
-#define LAB_MIN_VIEW_HEIGHT  60
+#if HAVE_XWAYLAND
+#include <wlr/xwayland.h>
+#endif
+
 #define LAB_FALLBACK_WIDTH  640
 #define LAB_FALLBACK_HEIGHT 480
 
+struct view *
+view_from_wlr_surface(struct wlr_surface *surface)
+{
+	assert(surface);
+	/*
+	 * TODO:
+	 * - find a way to get rid of xdg/xwayland-specific stuff
+	 * - look up root/toplevel surface if passed a subsurface?
+	 */
+	if (wlr_surface_is_xdg_surface(surface)) {
+		struct wlr_xdg_surface *xdg_surface =
+			wlr_xdg_surface_from_wlr_surface(surface);
+		if (xdg_surface) {
+			return xdg_surface->data;
+		}
+	}
+#if HAVE_XWAYLAND
+	if (wlr_surface_is_xwayland_surface(surface)) {
+		struct wlr_xwayland_surface *xsurface =
+			wlr_xwayland_surface_from_wlr_surface(surface);
+		if (xsurface) {
+			return xsurface->data;
+		}
+	}
+#endif
+	return NULL;
+}
+
+void
+view_query_free(struct view_query *query)
+{
+	wl_list_remove(&query->link);
+	free(query->identifier);
+	free(query->title);
+	free(query);
+}
+
+bool
+view_matches_query(struct view *view, struct view_query *query)
+{
+	bool match = true;
+	bool empty = true;
+
+	const char *identifier = view_get_string_prop(view, "app_id");
+	if (match && query->identifier) {
+		empty = false;
+		match &= match_glob(query->identifier, identifier);
+	}
+
+	const char *title = view_get_string_prop(view, "title");
+	if (match && query->title) {
+		empty = false;
+		match &= match_glob(query->title, title);
+	}
+
+	return !empty && match;
+}
+
 static bool
 matches_criteria(struct view *view, enum lab_view_criteria criteria)
 {
-	if (!view_isfocusable(view)) {
+	if (!view_is_focusable(view)) {
 		return false;
 	}
 	if (criteria & LAB_VIEW_CRITERIA_CURRENT_WORKSPACE) {
@@ -36,6 +100,16 @@ matches_criteria(struct view *view, enum lab_view_criteria criteria)
 			return false;
 		}
 	}
+	if (criteria & LAB_VIEW_CRITERIA_FULLSCREEN) {
+		if (!view->fullscreen) {
+			return false;
+		}
+	}
+	if (criteria & LAB_VIEW_CRITERIA_ALWAYS_ON_TOP) {
+		if (!view_is_always_on_top(view)) {
+			return false;
+		}
+	}
 	if (criteria & LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP) {
 		if (view_is_always_on_top(view)) {
 			return false;
@@ -80,13 +154,39 @@ view_array_append(struct server *server, struct wl_array *views,
 	}
 }
 
-bool
-view_isfocusable(struct view *view)
+enum view_wants_focus
+view_wants_focus(struct view *view)
 {
+	assert(view);
+	if (view->impl->wants_focus) {
+		return view->impl->wants_focus(view);
+	}
+	return VIEW_WANTS_FOCUS_ALWAYS;
+}
+
+bool
+view_is_focusable_from(struct view *view, struct wlr_surface *prev)
+{
+	assert(view);
 	if (!view->surface) {
 		return false;
 	}
-	return (view->mapped || view->minimized);
+	if (!view->mapped && !view->minimized) {
+		return false;
+	}
+	enum view_wants_focus wants_focus = view_wants_focus(view);
+	/*
+	 * Consider "offer focus" (Globally Active) views as focusable
+	 * only if another surface from the same application already had
+	 * focus. The goal is to allow focusing a parent window when a
+	 * dialog/popup is closed, but still avoid focusing standalone
+	 * panels/toolbars/notifications. Note that we are basically
+	 * guessing whether Globally Active views want focus, and will
+	 * probably be wrong some of the time.
+	 */
+	return (wants_focus == VIEW_WANTS_FOCUS_ALWAYS
+		|| (wants_focus == VIEW_WANTS_FOCUS_OFFER
+			&& prev && view_is_related(view, prev)));
 }
 
 /**
@@ -119,16 +219,7 @@ static struct wlr_box
 view_get_edge_snap_box(struct view *view, struct output *output,
 		enum view_edge edge)
 {
-	struct wlr_box usable = output_usable_area_in_layout_coords(output);
-	if (usable.height == output->wlr_output->height
-			&& output->wlr_output->scale != 1) {
-		usable.height /= output->wlr_output->scale;
-	}
-	if (usable.width == output->wlr_output->width
-			&& output->wlr_output->scale != 1) {
-		usable.width /= output->wlr_output->scale;
-	}
-
+	struct wlr_box usable = output_usable_area_scaled(output);
 	int x_offset = edge == VIEW_EDGE_RIGHT
 		? (usable.width + rc.gap) / 2 : rc.gap;
 	int y_offset = edge == VIEW_EDGE_DOWN
@@ -174,9 +265,10 @@ view_discover_output(struct view *view)
 		view->current.y + view->current.height / 2);
 }
 
-static void
-_view_set_activated(struct view *view, bool activated)
+void
+view_set_activated(struct view *view, bool activated)
 {
+	assert(view);
 	ssd_set_active(view->ssd, activated);
 	if (view->impl->set_activated) {
 		view->impl->set_activated(view, activated);
@@ -185,21 +277,17 @@ _view_set_activated(struct view *view, bool activated)
 		wlr_foreign_toplevel_handle_v1_set_activated(
 			view->toplevel.handle, activated);
 	}
-}
 
-void
-view_set_activated(struct view *view)
-{
-	assert(view);
-	struct view *last = view->server->focused_view;
-	if (last == view) {
-		return;
+	if (rc.kb_layout_per_window) {
+		if (!activated) {
+			/* Store configured keyboard layout per view */
+			view->keyboard_layout =
+				view->server->seat.keyboard_group->keyboard.modifiers.group;
+		} else {
+			/* Switch to previously stored keyboard layout */
+			keyboard_update_layout(&view->server->seat, view->keyboard_layout);
+		}
 	}
-	if (last) {
-		_view_set_activated(last, false);
-	}
-	_view_set_activated(view, true);
-	view->server->focused_view = view;
 }
 
 void
@@ -270,7 +358,8 @@ view_move_resize(struct view *view, struct wlr_box geo)
 void
 view_resize_relative(struct view *view, int left, int right, int top, int bottom)
 {
-	if (view->fullscreen || view->maximized) {
+	assert(view);
+	if (view->fullscreen || view->maximized != VIEW_AXIS_NONE) {
 		return;
 	}
 	struct wlr_box newgeo = view->pending;
@@ -285,10 +374,11 @@ view_resize_relative(struct view *view, int left, int right, int top, int bottom
 void
 view_move_relative(struct view *view, int x, int y)
 {
+	assert(view);
 	if (view->fullscreen) {
 		return;
 	}
-	view_maximize(view, false, /*store_natural_geometry*/ false);
+	view_maximize(view, VIEW_AXIS_NONE, /*store_natural_geometry*/ false);
 	if (view_is_tiled(view)) {
 		view_set_untiled(view);
 		view_restore_to(view, view->natural_geometry);
@@ -297,19 +387,108 @@ view_move_relative(struct view *view, int x, int y)
 }
 
 void
-view_adjust_size(struct view *view, int *w, int *h)
+view_move_to_cursor(struct view *view)
 {
 	assert(view);
 
-#if HAVE_XWAYLAND
-	if (xwayland_apply_size_hints(view, w, h)) {
-		/* We don't want to cap the size to keep the aspect ratio */
+	struct output *pending_output = output_nearest_to_cursor(view->server);
+	if (!output_is_usable(pending_output)) {
 		return;
 	}
-#endif
+	view_set_fullscreen(view, false);
+	view_maximize(view, VIEW_AXIS_NONE, /*store_natural_geometry*/ false);
+	if (view_is_tiled(view)) {
+		view_set_untiled(view);
+		view_restore_to(view, view->natural_geometry);
+	}
 
-	*w = MAX(*w, LAB_MIN_VIEW_WIDTH);
-	*h = MAX(*h, LAB_MIN_VIEW_HEIGHT);
+	struct border margin = ssd_thickness(view);
+	struct wlr_box geo = view->pending;
+	geo.width += margin.left + margin.right;
+	geo.height += margin.top + margin.bottom;
+
+	int x = view->server->seat.cursor->x - (geo.width / 2);
+	int y = view->server->seat.cursor->y - (geo.height / 2);
+
+	struct wlr_box usable = output_usable_area_in_layout_coords(pending_output);
+	if (x + geo.width > usable.x + usable.width) {
+		x = usable.x + usable.width - geo.width;
+	}
+	x = MAX(x, usable.x);
+
+	if (y + geo.height > usable.y + usable.height) {
+		y = usable.y + usable.height - geo.height;
+	}
+	y = MAX(y, usable.y);
+
+	x += margin.left;
+	y += margin.top;
+	view_move(view, x, y);
+}
+
+struct view_size_hints
+view_get_size_hints(struct view *view)
+{
+	assert(view);
+	if (view->impl->get_size_hints) {
+		return view->impl->get_size_hints(view);
+	}
+	return (struct view_size_hints){0};
+}
+
+static void
+substitute_nonzero(int *a, int *b)
+{
+	if (!(*a)) {
+		*a = *b;
+	} else if (!(*b)) {
+		*b = *a;
+	}
+}
+
+static int
+round_to_increment(int val, int base, int inc)
+{
+	if (base < 0 || inc <= 0) {
+		return val;
+	}
+	return base + (val - base + inc / 2) / inc * inc;
+}
+
+void
+view_adjust_size(struct view *view, int *w, int *h)
+{
+	assert(view);
+	struct view_size_hints hints = view_get_size_hints(view);
+
+	/*
+	 * "If a base size is not provided, the minimum size is to be
+	 * used in its place and vice versa." (ICCCM 4.1.2.3)
+	 */
+	substitute_nonzero(&hints.min_width, &hints.base_width);
+	substitute_nonzero(&hints.min_height, &hints.base_height);
+
+	/*
+	 * Snap width/height to requested size increments (if any).
+	 * Typically, terminal emulators use these to make sure that the
+	 * terminal is resized to a width/height evenly divisible by the
+	 * cell (character) size.
+	 */
+	*w = round_to_increment(*w, hints.base_width, hints.width_inc);
+	*h = round_to_increment(*h, hints.base_height, hints.height_inc);
+
+	/*
+	 * If a minimum width/height was not set, then use default.
+	 * This is currently always the case for xdg-shell views.
+	 */
+	if (hints.min_width < 1) {
+		hints.min_width = LAB_MIN_VIEW_WIDTH;
+	}
+	if (hints.min_height < 1) {
+		hints.min_height = LAB_MIN_VIEW_HEIGHT;
+	}
+	*w = MAX(*w, hints.min_width);
+	*h = MAX(*h, hints.min_height);
 }
 
 static void
@@ -329,18 +508,6 @@ _minimize(struct view *view, bool minimized)
 	view->minimized = minimized;
 	if (minimized) {
 		view->impl->unmap(view, /* client_request */ false);
-		_view_set_activated(view, false);
-		if (view == view->server->focused_view) {
-			/*
-			 * Prevents clearing the active view when
-			 * we don't actually have keyboard focus.
-			 *
-			 * This may happen when using a custom mouse
-			 * focus configuration or by using the foreign
-			 * toplevel protocol via some panel.
-			 */
-			view->server->focused_view = NULL;
-		}
 	} else {
 		view->impl->map(view);
 	}
@@ -370,6 +537,7 @@ minimize_sub_views(struct view *view, bool minimized)
 void
 view_minimize(struct view *view, bool minimized)
 {
+	assert(view);
 	/*
 	 * Minimize the root window first because some xwayland clients send a
 	 * request-unmap to sub-windows at this point (for example gimp and its
@@ -380,10 +548,11 @@ view_minimize(struct view *view, bool minimized)
 	minimize_sub_views(root, minimized);
 }
 
-static bool
+bool
 view_compute_centered_position(struct view *view, const struct wlr_box *ref,
 		int w, int h, int *x, int *y)
 {
+	assert(view);
 	if (w <= 0 || h <= 0) {
 		wlr_log(WLR_ERROR, "view has empty geometry, not centering");
 		return false;
@@ -419,6 +588,51 @@ view_compute_centered_position(struct view *view, const struct wlr_box *ref,
 	return true;
 }
 
+bool
+view_adjust_floating_geometry(struct view *view, struct wlr_box *geometry)
+{
+	assert(view);
+	if (!output_is_usable(view->output)) {
+		wlr_log(WLR_ERROR, "view has no output, not positioning");
+		return false;
+	}
+
+	if (window_rules_get_property(view, "fixedPosition") == LAB_PROP_TRUE) {
+		return false;
+	}
+
+	bool adjusted = false;
+	/*
+	 * First check whether the view is onscreen. For now, "onscreen"
+	 * is defined as even one pixel of the client area being visible.
+	 */
+	if (wlr_output_layout_intersects(view->server->output_layout,
+			NULL, geometry)) {
+		/*
+		 * If onscreen, then make sure the titlebar is also
+		 * visible (and not overlapping any panels/docks)
+		 */
+		struct border margin = ssd_get_margin(view->ssd);
+		struct wlr_box usable =
+			output_usable_area_in_layout_coords(view->output);
+
+		if (geometry->x < usable.x + margin.left) {
+			geometry->x = usable.x + margin.left;
+			adjusted = true;
+		}
+		if (geometry->y < usable.y + margin.top) {
+			geometry->y = usable.y + margin.top;
+			adjusted = true;
+		}
+	} else {
+		/* If offscreen, then just center the view */
+		adjusted = view_compute_centered_position(view, NULL,
+			geometry->width, geometry->height,
+			&geometry->x, &geometry->y);
+	}
+	return adjusted;
+}
+
 static void
 set_fallback_geometry(struct view *view)
 {
@@ -472,19 +686,9 @@ view_apply_natural_geometry(struct view *view)
 	assert(view);
 	assert(view_is_floating(view));
 
-	struct wlr_output_layout *layout = view->server->output_layout;
-	if (wlr_output_layout_intersects(layout, NULL, &view->natural_geometry)
-			|| wl_list_empty(&layout->outputs)) {
-		/* restore to original geometry */
-		view_move_resize(view, view->natural_geometry);
-	} else {
-		/* reposition if original geometry is offscreen */
-		struct wlr_box box = view->natural_geometry;
-		if (view_compute_centered_position(view, NULL, box.width,
-				box.height, &box.x, &box.y)) {
-			view_move_resize(view, box);
-		}
-	}
+	struct wlr_box geometry = view->natural_geometry;
+	view_adjust_floating_geometry(view, &geometry);
+	view_move_resize(view, geometry);
 }
 
 static void
@@ -589,7 +793,7 @@ static void
 view_apply_maximized_geometry(struct view *view)
 {
 	assert(view);
-	assert(view->maximized);
+	assert(view->maximized != VIEW_AXIS_NONE);
 	struct output *output = view->output;
 	assert(output_is_usable(output));
 
@@ -603,6 +807,23 @@ view_apply_maximized_geometry(struct view *view)
 		box.width /= output->wlr_output->scale;
 	}
 
+	/*
+	 * If one axis (horizontal or vertical) is unmaximized, it
+	 * should use the natural geometry. But if that geometry is not
+	 * on-screen on the output where the view is maximized, then
+	 * center the unmaximized axis.
+	 */
+	struct wlr_box natural = view->natural_geometry;
+	if (view->maximized != VIEW_AXIS_BOTH) {
+		struct wlr_box intersect;
+		wlr_box_intersection(&intersect, &box, &natural);
+		if (wlr_box_empty(&intersect)) {
+			view_compute_centered_position(view, NULL,
+				natural.width, natural.height,
+				&natural.x, &natural.y);
+		}
+	}
+
 	if (view->ssd_enabled) {
 		struct border border = ssd_thickness(view);
 		box.x += border.left;
@@ -610,6 +831,15 @@ view_apply_maximized_geometry(struct view *view)
 		box.width -= border.right + border.left;
 		box.height -= border.top + border.bottom;
 	}
+
+	if (view->maximized == VIEW_AXIS_VERTICAL) {
+		box.x = natural.x;
+		box.width = natural.width;
+	} else if (view->maximized == VIEW_AXIS_HORIZONTAL) {
+		box.y = natural.y;
+		box.height = natural.height;
+	}
+
 	view_move_resize(view, box);
 }
 
@@ -625,7 +855,7 @@ view_apply_special_geometry(struct view *view)
 
 	if (view->fullscreen) {
 		view_apply_fullscreen_geometry(view);
-	} else if (view->maximized) {
+	} else if (view->maximized != VIEW_AXIS_NONE) {
 		view_apply_maximized_geometry(view);
 	} else if (view->tiled) {
 		view_apply_tiled_geometry(view);
@@ -638,14 +868,14 @@ view_apply_special_geometry(struct view *view)
 
 /* For internal use only. Does not update geometry. */
 static void
-set_maximized(struct view *view, bool maximized)
+set_maximized(struct view *view, enum view_axis maximized)
 {
 	if (view->impl->maximize) {
-		view->impl->maximize(view, maximized);
+		view->impl->maximize(view, (maximized == VIEW_AXIS_BOTH));
 	}
 	if (view->toplevel.handle) {
 		wlr_foreign_toplevel_handle_v1_set_maximized(
-			view->toplevel.handle, maximized);
+			view->toplevel.handle, (maximized == VIEW_AXIS_BOTH));
 	}
 	view->maximized = maximized;
 
@@ -668,8 +898,8 @@ view_restore_to(struct view *view, struct wlr_box geometry)
 	if (view->fullscreen) {
 		return;
 	}
-	if (view->maximized) {
-		set_maximized(view, false);
+	if (view->maximized != VIEW_AXIS_NONE) {
+		set_maximized(view, VIEW_AXIS_NONE);
 	}
 	view_move_resize(view, geometry);
 }
@@ -677,15 +907,17 @@ view_restore_to(struct view *view, struct wlr_box geometry)
 bool
 view_is_tiled(struct view *view)
 {
-	return view && (view->tiled || view->tiled_region
+	assert(view);
+	return (view->tiled || view->tiled_region
 		|| view->tiled_region_evacuate);
 }
 
 bool
 view_is_floating(struct view *view)
 {
-	return view && !(view->fullscreen || view->maximized || view->tiled
-		|| view->tiled_region || view->tiled_region_evacuate);
+	assert(view);
+	return !(view->fullscreen || (view->maximized != VIEW_AXIS_NONE)
+		|| view_is_tiled(view));
 }
 
 /* Reset tiled state of view without changing geometry */
@@ -699,27 +931,28 @@ view_set_untiled(struct view *view)
 }
 
 void
-view_maximize(struct view *view, bool maximize, bool store_natural_geometry)
+view_maximize(struct view *view, enum view_axis axis,
+		bool store_natural_geometry)
 {
 	assert(view);
-	if (view->maximized == maximize) {
+	if (view->maximized == axis) {
 		return;
 	}
 	if (view->fullscreen) {
 		return;
 	}
-	if (maximize) {
+	if (axis != VIEW_AXIS_NONE) {
 		/*
 		 * Maximize via keybind or client request cancels
 		 * interactive move/resize since we can't move/resize
 		 * a maximized view.
 		 */
 		interactive_cancel(view);
-		if (store_natural_geometry) {
+		if (store_natural_geometry && view_is_floating(view)) {
 			view_store_natural_geometry(view);
 		}
 	}
-	set_maximized(view, maximize);
+	set_maximized(view, axis);
 	if (view_is_floating(view)) {
 		view_apply_natural_geometry(view);
 	} else {
@@ -728,11 +961,28 @@ view_maximize(struct view *view, bool maximize, bool store_natural_geometry)
 }
 
 void
-view_toggle_maximize(struct view *view)
+view_toggle_maximize(struct view *view, enum view_axis axis)
 {
 	assert(view);
-	view_maximize(view, !view->maximized,
-		/*store_natural_geometry*/ true);
+	switch (axis) {
+	case VIEW_AXIS_HORIZONTAL:
+	case VIEW_AXIS_VERTICAL:
+		/* Toggle one axis (XOR) */
+		view_maximize(view, view->maximized ^ axis,
+			/*store_natural_geometry*/ true);
+		break;
+	case VIEW_AXIS_BOTH:
+		/*
+		 * Maximize in both directions if unmaximized or partially
+		 * maximized, otherwise unmaximize.
+		 */
+		view_maximize(view, (view->maximized == VIEW_AXIS_BOTH) ?
+			VIEW_AXIS_NONE : VIEW_AXIS_BOTH,
+			/*store_natural_geometry*/ true);
+		break;
+	default:
+		break;
+	}
 }
 
 void
@@ -777,7 +1027,7 @@ view_toggle_always_on_top(struct view *view)
 	}
 }
 
-static bool
+bool
 view_is_always_on_bottom(struct view *view)
 {
 	assert(view);
@@ -882,9 +1132,7 @@ set_fullscreen(struct view *view, bool fullscreen)
 
 	/* Show fullscreen views above top-layer */
 	if (view->output) {
-		uint32_t top = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
-		wlr_scene_node_set_enabled(&view->output->layer_tree[top]->node,
-			!fullscreen);
+		desktop_update_top_layer_visiblity(view->server);
 	}
 }
 
@@ -943,9 +1191,9 @@ view_adjust_for_layout_change(struct view *view)
 		view_apply_natural_geometry(view);
 	} else {
 		/* reposition view if it's offscreen */
-		if (!wlr_output_layout_intersects(view->server->output_layout,
-				NULL, &view->pending)) {
-			view_center(view, NULL);
+		struct wlr_box geometry = view->pending;
+		if (view_adjust_floating_geometry(view, &geometry)) {
+			view_move_resize(view, geometry);
 		}
 	}
 	if (view->toplevel.handle) {
@@ -977,50 +1225,91 @@ view_on_output_destroy(struct view *view)
 }
 
 void
-view_move_to_edge(struct view *view, enum view_edge edge)
+view_move_to_edge(struct view *view, enum view_edge direction, bool snap_to_windows)
 {
 	assert(view);
-	struct output *output = view->output;
-	if (!output_is_usable(output)) {
+	if (!output_is_usable(view->output)) {
 		wlr_log(WLR_ERROR, "view has no output, not moving to edge");
 		return;
 	}
 
-	struct border margin = ssd_get_margin(view->ssd);
-	struct wlr_box usable = output_usable_area_in_layout_coords(output);
-	if (usable.height == output->wlr_output->height
-			&& output->wlr_output->scale != 1) {
-		usable.height /= output->wlr_output->scale;
-	}
-	if (usable.width == output->wlr_output->width
-			&& output->wlr_output->scale != 1) {
-		usable.width /= output->wlr_output->scale;
+	int dx = 0, dy = 0;
+	if (snap_to_windows) {
+		snap_vector_to_next_edge(view, direction, &dx, &dy);
+	} else {
+		struct border distance = snap_get_max_distance(view);
+		switch (direction) {
+		case VIEW_EDGE_LEFT:
+			dx = distance.left;
+			break;
+		case VIEW_EDGE_UP:
+			dy = distance.top;
+			break;
+		case VIEW_EDGE_RIGHT:
+			dx = distance.right;
+			break;
+		case VIEW_EDGE_DOWN:
+			dy = distance.bottom;
+			break;
+		default:
+			return;
+		}
 	}
 
-	int x = 0, y = 0;
-	switch (edge) {
-	case VIEW_EDGE_LEFT:
-		x = usable.x + margin.left + rc.gap;
-		y = view->pending.y;
-		break;
-	case VIEW_EDGE_UP:
-		x = view->pending.x;
-		y = usable.y + margin.top + rc.gap;
-		break;
-	case VIEW_EDGE_RIGHT:
-		x = usable.x + usable.width - view->pending.width
-			- margin.right - rc.gap;
-		y = view->pending.y;
-		break;
-	case VIEW_EDGE_DOWN:
-		x = view->pending.x;
-		y = usable.y + usable.height - view->pending.height
-			- margin.bottom - rc.gap;
-		break;
-	default:
+	view_move(view, view->pending.x + dx, view->pending.y + dy);
+}
+
+void
+view_grow_to_edge(struct view *view, enum view_edge direction)
+{
+	assert(view);
+	/* TODO: allow grow to edge if maximized along the other axis */
+	if (view->fullscreen || view->maximized != VIEW_AXIS_NONE) {
 		return;
 	}
-	view_move(view, x, y);
+	if (!output_is_usable(view->output)) {
+		wlr_log(WLR_ERROR, "view has no output, not growing view");
+		return;
+	}
+
+	struct wlr_box geo = view->pending;
+	snap_grow_to_next_edge(view, direction, &geo);
+	view_move_resize(view, geo);
+}
+
+void
+view_shrink_to_edge(struct view *view, enum view_edge direction)
+{
+	assert(view);
+	/* TODO: allow shrink to edge if maximized along the other axis */
+	if (view->fullscreen || view->maximized != VIEW_AXIS_NONE) {
+		return;
+	}
+	if (!output_is_usable(view->output)) {
+		wlr_log(WLR_ERROR, "view has no output, not shrinking view");
+		return;
+	}
+
+	struct wlr_box geo = view->pending;
+	snap_shrink_to_next_edge(view, direction, &geo);
+	view_move_resize(view, geo);
+}
+
+enum view_axis
+view_axis_parse(const char *direction)
+{
+	if (!direction) {
+		return VIEW_AXIS_NONE;
+	}
+	if (!strcasecmp(direction, "horizontal")) {
+		return VIEW_AXIS_HORIZONTAL;
+	} else if (!strcasecmp(direction, "vertical")) {
+		return VIEW_AXIS_VERTICAL;
+	} else if (!strcasecmp(direction, "both")) {
+		return VIEW_AXIS_BOTH;
+	} else {
+		return VIEW_AXIS_NONE;
+	}
 }
 
 enum view_edge
@@ -1045,7 +1334,8 @@ view_edge_parse(const char *direction)
 }
 
 void
-view_snap_to_edge(struct view *view, enum view_edge edge, bool store_natural_geometry)
+view_snap_to_edge(struct view *view, enum view_edge edge,
+			bool across_outputs, bool store_natural_geometry)
 {
 	assert(view);
 	if (view->fullscreen) {
@@ -1057,7 +1347,7 @@ view_snap_to_edge(struct view *view, enum view_edge edge, bool store_natural_geo
 		return;
 	}
 
-	if (view->tiled == edge && !view->maximized) {
+	if (across_outputs && view->tiled == edge && view->maximized == VIEW_AXIS_NONE) {
 		/* We are already tiled for this edge and thus should switch outputs */
 		struct wlr_output *new_output = NULL;
 		struct wlr_output *current_output = output->wlr_output;
@@ -1098,22 +1388,16 @@ view_snap_to_edge(struct view *view, enum view_edge edge, bool store_natural_geo
 			 * state because the window might have been moved away
 			 * (and thus got untiled) and then snapped back to the
 			 * original edge.
-			 *
-			 * TODO: The described pattern will cause another bug
-			 *       in multi monitor setups: it will snap the
-			 *       window to the inverted edge of the nearest
-			 *       output. This is the desired behavior when
-			 *       caused by a keybind but doesn't make sense
-			 *       when caused by mouse movement.
 			 */
 			view_apply_tiled_geometry(view);
 			return;
 		}
 	}
 
-	if (view->maximized) {
+	if (view->maximized != VIEW_AXIS_NONE) {
 		/* Unmaximize + keep using existing natural_geometry */
-		view_maximize(view, false, /*store_natural_geometry*/ false);
+		view_maximize(view, VIEW_AXIS_NONE,
+			/*store_natural_geometry*/ false);
 	} else if (store_natural_geometry) {
 		/* store current geometry as new natural_geometry */
 		view_store_natural_geometry(view);
@@ -1139,9 +1423,10 @@ view_snap_to_region(struct view *view, struct region *region,
 		return;
 	}
 
-	if (view->maximized) {
+	if (view->maximized != VIEW_AXIS_NONE) {
 		/* Unmaximize + keep using existing natural_geometry */
-		view_maximize(view, false, /*store_natural_geometry*/ false);
+		view_maximize(view, VIEW_AXIS_NONE,
+			/*store_natural_geometry*/ false);
 	} else if (store_natural_geometry) {
 		/* store current geometry as new natural_geometry */
 		view_store_natural_geometry(view);
@@ -1151,24 +1436,85 @@ view_snap_to_region(struct view *view, struct region *region,
 	view_apply_region_geometry(view);
 }
 
+static void
+for_each_subview(struct view *view, void (*action)(struct view *))
+{
+	struct wl_array subviews;
+	struct view **subview;
+
+	wl_array_init(&subviews);
+	view_append_children(view, &subviews);
+	wl_array_for_each(subview, &subviews) {
+		action(*subview);
+	}
+	wl_array_release(&subviews);
+}
+
+static void
+move_to_front(struct view *view)
+{
+	if (view->impl->move_to_front) {
+		view->impl->move_to_front(view);
+	}
+	view->server->last_raised_view = view;
+}
+
+static void
+move_to_back(struct view *view)
+{
+	if (view->impl->move_to_back) {
+		view->impl->move_to_back(view);
+	}
+	if (view == view->server->last_raised_view) {
+		view->server->last_raised_view = NULL;
+	}
+}
+
+/*
+ * In the view_move_to_{front,back} functions, a modal dialog is always
+ * shown above its parent window, and the two always move together, so
+ * other windows cannot come between them.
+ * This is consistent with GTK3/Qt5 applications on mutter and openbox.
+ */
 void
 view_move_to_front(struct view *view)
 {
 	assert(view);
-	if (view->impl->move_to_front) {
-		view->impl->move_to_front(view);
-		cursor_update_focus(view->server);
+	/*
+	 * This function is called often, generally on every mouse
+	 * button press (more often for focus-follows-mouse). Avoid
+	 * unnecessarily raising the same view over and over, or
+	 * attempting to raise a root view above its own sub-view.
+	 */
+	struct view *last = view->server->last_raised_view;
+	if (view == last || (last && view == view_get_root(last))) {
+		return;
 	}
+
+	struct view *root = view_get_root(view);
+	assert(root);
+
+	move_to_front(root);
+	for_each_subview(root, move_to_front);
+	/* make sure view is in front of other sub-views */
+	if (view != root) {
+		move_to_front(view);
+	}
+
+	cursor_update_focus(view->server);
 }
 
 void
 view_move_to_back(struct view *view)
 {
 	assert(view);
-	if (view->impl->move_to_back) {
-		view->impl->move_to_back(view);
-		cursor_update_focus(view->server);
-	}
+	struct view *root = view_get_root(view);
+	assert(root);
+
+	for_each_subview(root, move_to_back);
+	move_to_back(root);
+
+	cursor_update_focus(view->server);
 }
 
 struct view *
@@ -1190,6 +1536,17 @@ view_append_children(struct view *view, struct wl_array *children)
 	}
 }
 
+bool
+view_is_related(struct view *view, struct wlr_surface *surface)
+{
+	assert(view);
+	assert(surface);
+	if (view->impl->is_related) {
+		return view->impl->is_related(view, surface);
+	}
+	return false;
+}
+
 const char *
 view_get_string_prop(struct view *view, const char *prop)
 {
@@ -1235,36 +1592,23 @@ view_reload_ssd(struct view *view)
 	}
 }
 
-static void
-inhibit_keybinds(struct view *view, bool inhibit)
+void
+view_toggle_keybinds(struct view *view)
 {
-	assert(view->inhibits_keybinds != inhibit);
-
-	view->inhibits_keybinds = inhibit;
-	if (inhibit) {
+	assert(view);
+	view->inhibits_keybinds = !view->inhibits_keybinds;
+	if (view->inhibits_keybinds) {
 		view->server->seat.nr_inhibited_keybind_views++;
 	} else {
 		view->server->seat.nr_inhibited_keybind_views--;
 	}
 
 	if (view->ssd_enabled) {
-		ssd_enable_keybind_inhibit_indicator(view->ssd, inhibit);
+		ssd_enable_keybind_inhibit_indicator(view->ssd,
+			view->inhibits_keybinds);
 	}
 }
 
-bool
-view_inhibits_keybinds(struct view *view)
-{
-	return view && view->inhibits_keybinds;
-}
-
-void
-view_toggle_keybinds(struct view *view)
-{
-	assert(view);
-	inhibit_keybinds(view, !view->inhibits_keybinds);
-}
-
 void
 view_destroy(struct view *view)
 {
@@ -1299,6 +1643,10 @@ view_destroy(struct view *view)
 		need_cursor_update = true;
 	}
 
+	if (server->last_raised_view == view) {
+		server->last_raised_view = NULL;
+	}
+
 	if (server->seat.pressed.view == view) {
 		seat_reset_pressed(&server->seat);
 	}
@@ -1322,13 +1670,12 @@ view_destroy(struct view *view)
 
 	/*
 	 * The layer-shell top-layer is disabled when an application is running
-	 * in fullscreen mode, so if that's the case, we have to re-enable it
-	 * here.
+	 * in fullscreen mode, so if that's the case, we may have to re-enable
+	 * it here.
 	 */
 	if (view->fullscreen && view->output) {
-		uint32_t top = ZWLR_LAYER_SHELL_V1_LAYER_TOP;
-		wlr_scene_node_set_enabled(&view->output->layer_tree[top]->node,
-			true);
+		view->fullscreen = false;
+		desktop_update_top_layer_visiblity(server);
 	}
 
 	/* If we spawned a window menu, close it */
@@ -1345,4 +1692,3 @@ view_destroy(struct view *view)
 		cursor_update_focus(server);
 	}
 }
-
