@@ -7,14 +7,17 @@
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
+#include "config.h"
 #include "buffer.h"
 #include "common/font.h"
 #include "common/graphic-helpers.h"
 #include "common/list.h"
 #include "common/mem.h"
+#include "input/keyboard.h"
 #include "labwc.h"
 #include "view.h"
 #include "workspaces.h"
+#include "xwayland.h"
 
 /* Internal helpers */
 static size_t
@@ -255,8 +258,13 @@ workspaces_init(struct server *server)
 	}
 }
 
+/*
+ * update_focus should normally be set to true. It is set to false only
+ * when this function is called from desktop_focus_view(), in order to
+ * avoid unnecessary extra focus changes and possible recursion.
+ */
 void
-workspaces_switch_to(struct workspace *target)
+workspaces_switch_to(struct workspace *target, bool update_focus)
 {
 	assert(target);
 	struct server *server = target->server;
@@ -281,19 +289,28 @@ workspaces_switch_to(struct workspace *target)
 	 * Make sure we are focusing what the user sees.
 	 * Only refocus if the focus is not already on an always-on-top view.
 	 */
-	struct view *view = desktop_focused_view(server);
-	if (!view || !view_is_always_on_top(view)) {
-		desktop_focus_topmost_mapped_view(server);
+	if (update_focus) {
+		struct view *view = server->focused_view;
+		if (!view || !view_is_always_on_top(view)) {
+			desktop_focus_topmost_view(server);
+		}
 	}
 
 	/* And finally show the OSD */
 	_osd_show(server);
 
+#if HAVE_XWAYLAND
+	/* Ensure xwayland internal stacking order corresponds to the current workspace  */
+	xwayland_adjust_stacking_order(server);
+#endif
 	/*
 	 * Make sure we are not carrying around a
 	 * cursor image from the previous desktop
 	 */
 	cursor_update_focus(server);
+
+	/* Ensure that only currently visible fullscreen windows hide the top layer */
+	desktop_update_top_layer_visiblity(server);
 }
 
 void
@@ -333,6 +350,8 @@ workspaces_find(struct workspace *anchor, const char *name, bool wrap)
 				return target;
 			}
 		}
+	} else if (!strcasecmp(name, "current")) {
+		return anchor;
 	} else if (!strcasecmp(name, "last")) {
 		return anchor->server->workspace_last;
 	} else if (!strcasecmp(name, "left")) {
