@@ -1,5 +1,6 @@
 // SPDX-License-Identifier: GPL-2.0-only
 #include <assert.h>
+#include "common/macros.h"
 #include "common/mem.h"
 #include "decorations.h"
 #include "labwc.h"
@@ -9,6 +10,7 @@
 #include "window-rules.h"
 #include "workspaces.h"
 
+#define LAB_XDG_SHELL_VERSION (2)
 #define CONFIGURE_TIMEOUT_MS 100
 
 static struct xdg_toplevel_view *
@@ -38,7 +40,7 @@ xdg_toplevel_from_view(struct view *view)
 }
 
 static void
-handle_new_xdg_popup(struct wl_listener *listener, void *data)
+handle_new_popup(struct wl_listener *listener, void *data)
 {
 	struct xdg_toplevel_view *xdg_toplevel_view =
 		wl_container_of(listener, xdg_toplevel_view, new_popup);
@@ -88,6 +90,27 @@ handle_commit(struct wl_listener *listener, void *data)
 	struct wlr_box size;
 	wlr_xdg_surface_get_geometry(xdg_surface, &size);
 
+	/*
+	 * Qt applications occasionally fail to call set_window_geometry
+	 * after a configure request, but do correctly update the actual
+	 * surface extent. This results in a mismatch between the window
+	 * decorations (which follow the logical geometry) and the visual
+	 * size of the client area. As a workaround, we try to detect
+	 * this case and ignore the out-of-date window geometry.
+	 */
+	if (size.width != view->pending.width
+			|| size.height != view->pending.height) {
+		struct wlr_box extent;
+		wlr_surface_get_extends(xdg_surface->surface, &extent);
+		if (extent.width == view->pending.width
+				&& extent.height == view->pending.height) {
+			wlr_log(WLR_DEBUG, "window geometry for client (%s) "
+				"appears to be incorrect - ignoring",
+				view_get_string_prop(view, "app_id"));
+			size = extent; /* Use surface extent instead */
+		}
+	}
+
 	struct wlr_box *current = &view->current;
 	bool update_required = current->width != size.width
 		|| current->height != size.height;
@@ -224,7 +247,8 @@ handle_request_maximize(struct wl_listener *listener, void *data)
 	if (!view->mapped && !view->output) {
 		view_set_output(view, output_nearest_to_cursor(view->server));
 	}
-	view_maximize(view, xdg_toplevel_from_view(view)->requested.maximized,
+	bool maximized = xdg_toplevel_from_view(view)->requested.maximized;
+	view_maximize(view, maximized ? VIEW_AXIS_BOTH : VIEW_AXIS_NONE,
 		/*store_natural_geometry*/ true);
 }
 
@@ -339,28 +363,6 @@ xdg_toplevel_view_get_root(struct view *view)
 	return (struct view *)surface->data;
 }
 
-/*
- * In the view_move_to_{front,back} functions, a modal dialog is always shown
- * above its parent window, and the two always move together, so other window
- * cannot come between them.
- * This is consistent with GTK3/Qt5 applications on mutter and openbox.
- */
-static void
-xdg_toplevel_view_move_to_front(struct view *view)
-{
-	struct view *root = xdg_toplevel_view_get_root(view);
-	view_impl_move_to_front(root);
-	view_impl_move_sub_views(root, LAB_TO_FRONT);
-}
-
-static void
-xdg_toplevel_view_move_to_back(struct view *view)
-{
-	struct view *root = xdg_toplevel_view_get_root(view);
-	view_impl_move_sub_views(root, LAB_TO_BACK);
-	view_impl_move_to_back(root);
-}
-
 static void
 xdg_toplevel_view_append_children(struct view *self, struct wl_array *children)
 {
@@ -523,12 +525,10 @@ xdg_toplevel_view_map(struct view *view)
 			position_xdg_toplevel_view(view);
 		}
 
-		if (!view->fullscreen && requested->fullscreen) {
-			set_fullscreen_from_request(view, requested);
-		} else if (!view->maximized && requested->maximized) {
-			view_maximize(view, true,
-				/*store_natural_geometry*/ true);
-		}
+		set_fullscreen_from_request(view, requested);
+		view_maximize(view, requested->maximized ?
+			VIEW_AXIS_BOTH : VIEW_AXIS_NONE,
+			/*store_natural_geometry*/ true);
 
 		/*
 		 * Set initial "current" position directly before
@@ -554,7 +554,7 @@ xdg_toplevel_view_unmap(struct view *view, bool client_request)
 		view->mapped = false;
 		wlr_scene_node_set_enabled(&view->scene_tree->node, false);
 		wl_list_remove(&view->commit.link);
-		desktop_focus_topmost_mapped_view(view->server);
+		view_impl_unmap(view);
 	}
 }
 
@@ -568,13 +568,13 @@ static const struct view_impl xdg_toplevel_view_impl = {
 	.unmap = xdg_toplevel_view_unmap,
 	.maximize = xdg_toplevel_view_maximize,
 	.minimize = xdg_toplevel_view_minimize,
-	.move_to_front = xdg_toplevel_view_move_to_front,
-	.move_to_back = xdg_toplevel_view_move_to_back,
+	.move_to_front = view_impl_move_to_front,
+	.move_to_back = view_impl_move_to_back,
 	.get_root = xdg_toplevel_view_get_root,
 	.append_children = xdg_toplevel_view_append_children,
 };
 
-void
+static void
 xdg_activation_handle_request(struct wl_listener *listener, void *data)
 {
 	const struct wlr_xdg_activation_v1_request_activate_event *event = data;
@@ -607,16 +607,8 @@ xdg_activation_handle_request(struct wl_listener *listener, void *data)
 		return;
 	}
 
-	/*
-	 * TODO: This is the exact same code as used in foreign.c.
-	 *       Refactor it into a public helper function somewhere.
-	 */
 	wlr_log(WLR_DEBUG, "Activating surface");
-	if (view->workspace != view->server->workspace_current) {
-		workspaces_switch_to(view->workspace);
-	}
-	desktop_focus_and_activate_view(&view->server->seat, view);
-	view_move_to_front(view);
+	desktop_focus_view(view, /*raise*/ true);
 }
 
 /*
@@ -626,7 +618,7 @@ xdg_activation_handle_request(struct wl_listener *listener, void *data)
  *   - wlr_surface->data = scene_tree
  *     to help the popups find their parent nodes
  */
-void
+static void
 xdg_surface_new(struct wl_listener *listener, void *data)
 {
 	struct server *server =
@@ -696,34 +688,44 @@ xdg_surface_new(struct wl_listener *listener, void *data)
 	/* In support of xdg popups */
 	xdg_surface->surface->data = tree;
 
-	view->map.notify = handle_map;
-	wl_signal_add(&xdg_surface->events.map, &view->map);
-	view->unmap.notify = handle_unmap;
-	wl_signal_add(&xdg_surface->events.unmap, &view->unmap);
-	view->destroy.notify = handle_destroy;
-	wl_signal_add(&xdg_surface->events.destroy, &view->destroy);
+	CONNECT_SIGNAL(xdg_surface, view, map);
+	CONNECT_SIGNAL(xdg_surface, view, unmap);
+	CONNECT_SIGNAL(xdg_surface, view, destroy);
 
 	struct wlr_xdg_toplevel *toplevel = xdg_surface->toplevel;
-	view->request_move.notify = handle_request_move;
-	wl_signal_add(&toplevel->events.request_move, &view->request_move);
-	view->request_resize.notify = handle_request_resize;
-	wl_signal_add(&toplevel->events.request_resize, &view->request_resize);
-	view->request_minimize.notify = handle_request_minimize;
-	wl_signal_add(&toplevel->events.request_minimize, &view->request_minimize);
-	view->request_maximize.notify = handle_request_maximize;
-	wl_signal_add(&toplevel->events.request_maximize, &view->request_maximize);
-	view->request_fullscreen.notify = handle_request_fullscreen;
-	wl_signal_add(&toplevel->events.request_fullscreen, &view->request_fullscreen);
-
-	view->set_title.notify = handle_set_title;
-	wl_signal_add(&toplevel->events.set_title, &view->set_title);
+	CONNECT_SIGNAL(toplevel, view, request_move);
+	CONNECT_SIGNAL(toplevel, view, request_resize);
+	CONNECT_SIGNAL(toplevel, view, request_minimize);
+	CONNECT_SIGNAL(toplevel, view, request_maximize);
+	CONNECT_SIGNAL(toplevel, view, request_fullscreen);
+	CONNECT_SIGNAL(toplevel, view, set_title);
 
 	/* Events specific to XDG toplevel views */
-	xdg_toplevel_view->set_app_id.notify = handle_set_app_id;
-	wl_signal_add(&toplevel->events.set_app_id, &xdg_toplevel_view->set_app_id);
-
-	xdg_toplevel_view->new_popup.notify = handle_new_xdg_popup;
-	wl_signal_add(&xdg_surface->events.new_popup, &xdg_toplevel_view->new_popup);
+	CONNECT_SIGNAL(toplevel, xdg_toplevel_view, set_app_id);
+	CONNECT_SIGNAL(xdg_surface, xdg_toplevel_view, new_popup);
 
 	wl_list_insert(&server->views, &view->link);
 }
+
+void
+xdg_shell_init(struct server *server)
+{
+	server->xdg_shell = wlr_xdg_shell_create(server->wl_display,
+		LAB_XDG_SHELL_VERSION);
+	if (!server->xdg_shell) {
+		wlr_log(WLR_ERROR, "unable to create the XDG shell interface");
+		exit(EXIT_FAILURE);
+	}
+	server->new_xdg_surface.notify = xdg_surface_new;
+	wl_signal_add(&server->xdg_shell->events.new_surface, &server->new_xdg_surface);
+
+	server->xdg_activation = wlr_xdg_activation_v1_create(server->wl_display);
+	if (!server->xdg_activation) {
+		wlr_log(WLR_ERROR, "unable to create xdg_activation interface");
+		exit(EXIT_FAILURE);
+	}
+	server->xdg_activation_request.notify = xdg_activation_handle_request;
+	wl_signal_add(&server->xdg_activation->events.request_activate,
+		&server->xdg_activation_request);
+}
+
