@@ -3,6 +3,8 @@
 #include <assert.h>
 #include <stdlib.h>
 #include <wlr/xwayland.h>
+#include "common/array.h"
+#include "common/macros.h"
 #include "common/mem.h"
 #include "labwc.h"
 #include "node.h"
@@ -13,56 +15,90 @@
 #include "workspaces.h"
 #include "xwayland.h"
 
-static int
-round_to_increment(int val, int base, int inc)
+static struct view_size_hints
+xwayland_view_get_size_hints(struct view *view)
 {
-	if (base < 0 || inc <= 0) {
-		return val;
+	xcb_size_hints_t *hints = xwayland_surface_from_view(view)->size_hints;
+	if (!hints) {
+		return (struct view_size_hints){0};
 	}
-	return base + (val - base + inc / 2) / inc * inc;
+	return (struct view_size_hints){
+		.min_width = hints->min_width,
+		.min_height = hints->min_height,
+		.width_inc = hints->width_inc,
+		.height_inc = hints->height_inc,
+		.base_width = hints->base_width,
+		.base_height = hints->base_height,
+	};
 }
 
-bool
-xwayland_apply_size_hints(struct view *view, int *w, int *h)
+static enum view_wants_focus
+xwayland_view_wants_focus(struct view *view)
 {
-	assert(view);
-	if (view->type == LAB_XWAYLAND_VIEW) {
-		xcb_size_hints_t *hints =
-			xwayland_surface_from_view(view)->size_hints;
+	struct wlr_xwayland_surface *xsurface =
+		xwayland_surface_from_view(view);
 
+	switch (wlr_xwayland_icccm_input_model(xsurface)) {
+	/*
+	 * Abbreviated from ICCCM section 4.1.7 (Input Focus):
+	 *
+	 * Passive Input - The client expects keyboard input but never
+	 * explicitly sets the input focus.
+	 * Locally Active Input - The client expects keyboard input and
+	 * explicitly sets the input focus, but it only does so when one
+	 * of its windows already has the focus.
+	 *
+	 * Passive and Locally Active clients set the input field of
+	 * WM_HINTS to True, which indicates that they require window
+	 * manager assistance in acquiring the input focus.
+	 */
+	case WLR_ICCCM_INPUT_MODEL_PASSIVE:
+	case WLR_ICCCM_INPUT_MODEL_LOCAL:
+		return VIEW_WANTS_FOCUS_ALWAYS;
+
+	/*
+	 * Globally Active Input - The client expects keyboard input and
+	 * explicitly sets the input focus, even when it is in windows
+	 * the client does not own. ... It wants to prevent the window
+	 * manager from setting the input focus to any of its windows
+	 * [because it may or may not want focus].
+	 *
+	 * Globally Active client windows may receive a WM_TAKE_FOCUS
+	 * message from the window manager. If they want the focus, they
+	 * should respond with a SetInputFocus request.
+	 *
+	 * [Currently, labwc does not fully support clients voluntarily
+	 * taking focus via the WM_TAKE_FOCUS + SetInputFocus mechanism.
+	 * Instead, we try to guess whether the window wants focus based
+	 * on some heuristics -- see below.]
+	 */
+	case WLR_ICCCM_INPUT_MODEL_GLOBAL:
 		/*
-		 * Honor size increments from WM_SIZE_HINTS. Typically, X11
-		 * terminal emulators will use WM_SIZE_HINTS to make sure that
-		 * the terminal is resized to a width/height evenly divisible by
-		 * the cell (character) size.
+		 * Assume the window does want focus if it wants window
+		 * decorations (according to _MOTIF_WM_HINTS). This is
+		 * a stop-gap fix to ensure that various applications
+		 * (mainly Java-based ones such as IntelliJ IDEA) get
+		 * focus normally and appear in the window switcher. It
+		 * would be better to match based on _NET_WM_WINDOW_TYPE
+		 * instead, but that property isn't currently available
+		 * through wlroots API.
 		 */
-		if (hints) {
-			*w = round_to_increment(*w, hints->base_width,
-				hints->width_inc);
-			*h = round_to_increment(*h, hints->base_height,
-				hints->height_inc);
+		return (xsurface->decorations ==
+			WLR_XWAYLAND_SURFACE_DECORATIONS_ALL) ?
+			VIEW_WANTS_FOCUS_ALWAYS : VIEW_WANTS_FOCUS_OFFER;
 
-			*w = MAX(*w, MAX(1, hints->min_width));
-			*h = MAX(*h, MAX(1, hints->min_height));
-			return true;
-		}
+	/*
+	 * No Input - The client never expects keyboard input.
+	 *
+	 * No Input and Globally Active clients set the input field to
+	 * False, which requests that the window manager not set the
+	 * input focus to their top-level window.
+	 */
+	case WLR_ICCCM_INPUT_MODEL_NONE:
+		break;
 	}
-	return false;
-}
 
-static void
-xwayland_view_fill_size_hints(struct view *view, struct wlr_box *box)
-{
-	if (view->type == LAB_XWAYLAND_VIEW) {
-		xcb_size_hints_t *hints = xwayland_surface_from_view(view)->size_hints;
-		if (hints) {
-			box->width = hints->width_inc;
-			box->height = hints->height_inc;
-			return;
-		}
-	}
-	box->width = 0;
-	box->height = 0;
+	return VIEW_WANTS_FOCUS_NEVER;
 }
 
 static struct wlr_xwayland_surface *
@@ -232,7 +268,8 @@ handle_destroy(struct wl_listener *listener, void *data)
 	/*
 	 * Break view <-> xsurface association.  Note that the xsurface
 	 * may not actually be destroyed at this point; it may become an
-	 * "unmanaged" surface instead.
+	 * "unmanaged" surface instead (in that case it is important
+	 * that xsurface->data not point to the destroyed view).
 	 */
 	xwayland_view->xwayland_surface->data = NULL;
 	xwayland_view->xwayland_surface = NULL;
@@ -240,9 +277,9 @@ handle_destroy(struct wl_listener *listener, void *data)
 	/* Remove XWayland view specific listeners */
 	wl_list_remove(&xwayland_view->request_activate.link);
 	wl_list_remove(&xwayland_view->request_configure.link);
-	wl_list_remove(&xwayland_view->set_app_id.link);
+	wl_list_remove(&xwayland_view->set_class.link);
 	wl_list_remove(&xwayland_view->set_decorations.link);
-	wl_list_remove(&xwayland_view->override_redirect.link);
+	wl_list_remove(&xwayland_view->set_override_redirect.link);
 
 	view_destroy(view);
 }
@@ -291,8 +328,7 @@ handle_request_activate(struct wl_listener *listener, void *data)
 		return;
 	}
 
-	desktop_focus_and_activate_view(&view->server->seat, view);
-	view_move_to_front(view);
+	desktop_focus_view(view, /*raise*/ true);
 }
 
 static void
@@ -316,7 +352,7 @@ handle_request_maximize(struct wl_listener *listener, void *data)
 		view_set_decorations(view,
 			want_deco(xwayland_surface_from_view(view)));
 	}
-	view_toggle_maximize(view);
+	view_toggle_maximize(view, VIEW_AXIS_BOTH);
 }
 
 static void
@@ -341,7 +377,7 @@ static void
 handle_set_class(struct wl_listener *listener, void *data)
 {
 	struct xwayland_view *xwayland_view =
-		wl_container_of(listener, xwayland_view, set_app_id);
+		wl_container_of(listener, xwayland_view, set_class);
 	struct view *view = &xwayland_view->base;
 	view_update_app_id(view);
 }
@@ -390,10 +426,10 @@ handle_set_decorations(struct wl_listener *listener, void *data)
 }
 
 static void
-handle_override_redirect(struct wl_listener *listener, void *data)
+handle_set_override_redirect(struct wl_listener *listener, void *data)
 {
 	struct xwayland_view *xwayland_view =
-		wl_container_of(listener, xwayland_view, override_redirect);
+		wl_container_of(listener, xwayland_view, set_override_redirect);
 	struct view *view = &xwayland_view->base;
 	struct wlr_xwayland_surface *xsurface = xwayland_view->xwayland_surface;
 
@@ -418,26 +454,33 @@ set_initial_position(struct view *view,
 			XCB_ICCCM_SIZE_HINT_P_POSITION));
 
 	if (has_position) {
-		/* Just make sure the view is on-screen */
-		view_adjust_for_layout_change(view);
+		/*
+		 * Make sure a floating view is onscreen. For a
+		 * maximized/fullscreen view, do nothing; if it is
+		 * unmaximized/leaves fullscreen later, we will make
+		 * sure it is on-screen at that point.
+		 */
+		if (view_is_floating(view)) {
+			view_adjust_for_layout_change(view);
+		}
 	} else {
-		view_center(view, NULL);
+		if (view_is_floating(view)) {
+			view_center(view, NULL);
+		} else {
+			/*
+			 * View is maximized/fullscreen. Center the
+			 * stored natural geometry without actually
+			 * moving the view.
+			 */
+			view_compute_centered_position(view, NULL,
+				view->natural_geometry.width,
+				view->natural_geometry.height,
+				&view->natural_geometry.x,
+				&view->natural_geometry.y);
+		}
 	}
 }
 
-static void
-top_left_edge_boundary_check(struct view *view)
-{
-	struct wlr_box deco = ssd_max_extents(view);
-	if (deco.x < 0) {
-		view->current.x -= deco.x;
-	}
-	if (deco.y < 0) {
-		view->current.y -= deco.y;
-	}
-	view->impl->configure(view, view->current);
-}
-
 static void
 init_foreign_toplevel(struct view *view)
 {
@@ -474,9 +517,6 @@ xwayland_view_map(struct view *view)
 	view->mapped = true;
 	ensure_initial_geometry_and_output(view);
 	wlr_scene_node_set_enabled(&view->scene_tree->node, true);
-	if (!view->fullscreen && xwayland_surface->fullscreen) {
-		view_set_fullscreen(view, true);
-	}
 
 	if (view->surface != xwayland_surface->surface) {
 		if (view->surface) {
@@ -499,17 +539,36 @@ xwayland_view_map(struct view *view)
 		view->scene_node = &tree->node;
 	}
 
+	/*
+	 * Per the Extended Window Manager Hints (EWMH) spec: "The Window
+	 * Manager SHOULD honor _NET_WM_STATE whenever a withdrawn window
+	 * requests to be mapped."
+	 *
+	 * The following order of operations is intended to reduce the
+	 * number of resize (Configure) events:
+	 *   1. set fullscreen state
+	 *   2. set decorations (depends on fullscreen state)
+	 *   3. set maximized (geometry depends on decorations)
+	 */
+	view_set_fullscreen(view, xwayland_surface->fullscreen);
+	if (!view->been_mapped) {
+		view_set_decorations(view, want_deco(xwayland_surface));
+	}
+	enum view_axis axis = VIEW_AXIS_NONE;
+	if (xwayland_surface->maximized_horz) {
+		axis |= VIEW_AXIS_HORIZONTAL;
+	}
+	if (xwayland_surface->maximized_vert) {
+		axis |= VIEW_AXIS_VERTICAL;
+	}
+	view_maximize(view, axis, /*store_natural_geometry*/ true);
+
 	if (!view->toplevel.handle) {
 		init_foreign_toplevel(view);
 	}
 
 	if (!view->been_mapped) {
-		view_set_decorations(view, want_deco(xwayland_surface));
-
-		if (view_is_floating(view)) {
-			set_initial_position(view, xwayland_surface);
-		}
-
+		set_initial_position(view, xwayland_surface);
 		/*
 		 * When mapping the view for the first time, visual
 		 * artifacts are reduced if we display it immediately at
@@ -520,10 +579,6 @@ xwayland_view_map(struct view *view)
 		view_moved(view);
 	}
 
-	if (view->ssd_enabled && view_is_floating(view)) {
-		top_left_edge_boundary_check(view);
-	}
-
 	/* Add commit here, as xwayland map/unmap can change the wlr_surface */
 	wl_signal_add(&xwayland_surface->surface->events.commit, &view->commit);
 	view->commit.notify = handle_commit;
@@ -541,7 +596,7 @@ xwayland_view_unmap(struct view *view, bool client_request)
 	view->mapped = false;
 	wl_list_remove(&view->commit.link);
 	wlr_scene_node_set_enabled(&view->scene_tree->node, false);
-	desktop_focus_topmost_mapped_view(view->server);
+	view_impl_unmap(view);
 
 	/*
 	 * If the view was explicitly unmapped by the client (rather
@@ -569,27 +624,43 @@ xwayland_view_minimize(struct view *view, bool minimized)
 		minimized);
 }
 
-static struct view *
-xwayland_view_get_root(struct view *view)
-{
-	struct wlr_xwayland_surface *root = top_parent_of(view);
-	return (struct view *)root->data;
-}
-
 static void
 xwayland_view_move_to_front(struct view *view)
 {
-	struct view *root = xwayland_view_get_root(view);
-	view_impl_move_to_front(root);
-	view_impl_move_sub_views(root, LAB_TO_FRONT);
+	view_impl_move_to_front(view);
+	/*
+	 * Update XWayland stacking order.
+	 *
+	 * FIXME: it would be better to restack above the next lower
+	 * view, rather than on top of all other surfaces. Restacking
+	 * the unmanaged surfaces afterward is ugly and still doesn't
+	 * account for always-on-top views.
+	 */
+	wlr_xwayland_surface_restack(xwayland_surface_from_view(view),
+		NULL, XCB_STACK_MODE_ABOVE);
+	/* Restack unmanaged surfaces on top */
+	struct wl_list *list = &view->server->unmanaged_surfaces;
+	struct xwayland_unmanaged *u;
+	wl_list_for_each(u, list, link) {
+		wlr_xwayland_surface_restack(u->xwayland_surface,
+			NULL, XCB_STACK_MODE_ABOVE);
+	}
 }
 
 static void
 xwayland_view_move_to_back(struct view *view)
 {
-	struct view *root = xwayland_view_get_root(view);
-	view_impl_move_sub_views(root, LAB_TO_BACK);
-	view_impl_move_to_back(root);
+	view_impl_move_to_back(view);
+	/* Update XWayland stacking order */
+	wlr_xwayland_surface_restack(xwayland_surface_from_view(view),
+		NULL, XCB_STACK_MODE_BELOW);
+}
+
+static struct view *
+xwayland_view_get_root(struct view *view)
+{
+	struct wlr_xwayland_surface *root = top_parent_of(view);
+	return (struct view *)root->data;
 }
 
 static void
@@ -624,6 +695,18 @@ xwayland_view_append_children(struct view *self, struct wl_array *children)
 	}
 }
 
+static bool
+xwayland_view_is_related(struct view *view, struct wlr_surface *surface)
+{
+	struct wlr_xwayland_surface *xsurface =
+		xwayland_surface_from_view(view);
+	struct wlr_xwayland_surface *xsurface2 =
+		wlr_surface_is_xwayland_surface(surface) ?
+		wlr_xwayland_surface_from_wlr_surface(surface) : NULL;
+
+	return (xsurface2 && xsurface2->pid == xsurface->pid);
+}
+
 static void
 xwayland_view_set_activated(struct view *view, bool activated)
 {
@@ -635,17 +718,6 @@ xwayland_view_set_activated(struct view *view, bool activated)
 	}
 
 	wlr_xwayland_surface_activate(xwayland_surface, activated);
-	if (activated) {
-		wlr_xwayland_surface_restack(xwayland_surface,
-			NULL, XCB_STACK_MODE_ABOVE);
-		/* Restack unmanaged surfaces on top */
-		struct xwayland_unmanaged *u;
-		struct wl_list *list = &view->server->unmanaged_surfaces;
-		wl_list_for_each(u, list, link) {
-			wlr_xwayland_surface_restack(u->xwayland_surface,
-				NULL, XCB_STACK_MODE_ABOVE);
-		}
-	}
 }
 
 static void
@@ -669,7 +741,9 @@ static const struct view_impl xwayland_view_impl = {
 	.move_to_back = xwayland_view_move_to_back,
 	.get_root = xwayland_view_get_root,
 	.append_children = xwayland_view_append_children,
-	.fill_size_hints = xwayland_view_fill_size_hints,
+	.is_related = xwayland_view_is_related,
+	.get_size_hints = xwayland_view_get_size_hints,
+	.wants_focus = xwayland_view_wants_focus,
 };
 
 void
@@ -697,41 +771,22 @@ xwayland_view_create(struct server *server,
 	view->scene_tree = wlr_scene_tree_create(view->workspace->tree);
 	node_descriptor_create(&view->scene_tree->node, LAB_NODE_DESC_VIEW, view);
 
-	view->map.notify = handle_map;
-	wl_signal_add(&xsurface->events.map, &view->map);
-	view->unmap.notify = handle_unmap;
-	wl_signal_add(&xsurface->events.unmap, &view->unmap);
-	view->destroy.notify = handle_destroy;
-	wl_signal_add(&xsurface->events.destroy, &view->destroy);
-	view->request_minimize.notify = handle_request_minimize;
-	wl_signal_add(&xsurface->events.request_minimize, &view->request_minimize);
-	view->request_maximize.notify = handle_request_maximize;
-	wl_signal_add(&xsurface->events.request_maximize, &view->request_maximize);
-	view->request_fullscreen.notify = handle_request_fullscreen;
-	wl_signal_add(&xsurface->events.request_fullscreen, &view->request_fullscreen);
-	view->request_move.notify = handle_request_move;
-	wl_signal_add(&xsurface->events.request_move, &view->request_move);
-	view->request_resize.notify = handle_request_resize;
-	wl_signal_add(&xsurface->events.request_resize, &view->request_resize);
-
-	view->set_title.notify = handle_set_title;
-	wl_signal_add(&xsurface->events.set_title, &view->set_title);
+	CONNECT_SIGNAL(xsurface, view, map);
+	CONNECT_SIGNAL(xsurface, view, unmap);
+	CONNECT_SIGNAL(xsurface, view, destroy);
+	CONNECT_SIGNAL(xsurface, view, request_minimize);
+	CONNECT_SIGNAL(xsurface, view, request_maximize);
+	CONNECT_SIGNAL(xsurface, view, request_fullscreen);
+	CONNECT_SIGNAL(xsurface, view, request_move);
+	CONNECT_SIGNAL(xsurface, view, request_resize);
+	CONNECT_SIGNAL(xsurface, view, set_title);
 
 	/* Events specific to XWayland views */
-	xwayland_view->request_activate.notify = handle_request_activate;
-	wl_signal_add(&xsurface->events.request_activate, &xwayland_view->request_activate);
-
-	xwayland_view->request_configure.notify = handle_request_configure;
-	wl_signal_add(&xsurface->events.request_configure, &xwayland_view->request_configure);
-
-	xwayland_view->set_app_id.notify = handle_set_class;
-	wl_signal_add(&xsurface->events.set_class, &xwayland_view->set_app_id);
-
-	xwayland_view->set_decorations.notify = handle_set_decorations;
-	wl_signal_add(&xsurface->events.set_decorations, &xwayland_view->set_decorations);
-
-	xwayland_view->override_redirect.notify = handle_override_redirect;
-	wl_signal_add(&xsurface->events.set_override_redirect, &xwayland_view->override_redirect);
+	CONNECT_SIGNAL(xsurface, xwayland_view, request_activate);
+	CONNECT_SIGNAL(xsurface, xwayland_view, request_configure);
+	CONNECT_SIGNAL(xsurface, xwayland_view, set_class);
+	CONNECT_SIGNAL(xsurface, xwayland_view, set_decorations);
+	CONNECT_SIGNAL(xsurface, xwayland_view, set_override_redirect);
 
 	wl_list_insert(&view->server->views, &view->link);
 
@@ -803,6 +858,46 @@ xwayland_server_init(struct server *server, struct wlr_compositor *compositor)
 	}
 }
 
+/*
+ * Until we expose the workspaces to xwayland we need a way to
+ * ensure that xwayland views on the current workspace are always
+ * stacked above xwayland views on other workspaces.
+ *
+ * If we fail to do so, issues arise in scenarios where we change
+ * the mouse focus but do not change the (xwayland) stacking order.
+ *
+ * Reproducer:
+ * - open at least two xwayland windows which allow scrolling
+ *   (some X11 terminal with 'man man' for example)
+ * - switch to another workspace, open another xwayland
+ *   window which allows scrolling and maximize it
+ * - switch back to the previous workspace with the two windows
+ * - move the mouse to the xwayland window that does *not* have focus
+ * - start scrolling
+ * - all scroll events should end up on the maximized window on the other workspace
+ */
+void
+xwayland_adjust_stacking_order(struct server *server)
+{
+	struct view **view;
+	struct wl_array views;
+
+	wl_array_init(&views);
+	view_array_append(server, &views, LAB_VIEW_CRITERIA_ALWAYS_ON_TOP);
+	view_array_append(server, &views, LAB_VIEW_CRITERIA_CURRENT_WORKSPACE
+		| LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP);
+
+	/*
+	 * view_array_append() provides top-most windows
+	 * first so we simply reverse the iteration here
+	 */
+	wl_array_for_each_reverse(view, &views) {
+		view_move_to_front(*view);
+	}
+
+	wl_array_release(&views);
+}
+
 void
 xwayland_server_finish(struct server *server)
 {
